from __future__ import annotations

# Minimal ABIs for Aave v3 pool operations + UI periphery helpers.

POOL_ABI = [
    {
        "name": "supply",
        "type": "function",
        "stateMutability": "nonpayable",
        "inputs": [
            {"name": "asset", "type": "address"},
            {"name": "amount", "type": "uint256"},
            {"name": "onBehalfOf", "type": "address"},
            {"name": "referralCode", "type": "uint16"},
        ],
        "outputs": [],
    },
    {
        "name": "withdraw",
        "type": "function",
        "stateMutability": "nonpayable",
        "inputs": [
            {"name": "asset", "type": "address"},
            {"name": "amount", "type": "uint256"},
            {"name": "to", "type": "address"},
        ],
        "outputs": [{"name": "", "type": "uint256"}],
    },
    {
        "name": "borrow",
        "type": "function",
        "stateMutability": "nonpayable",
        "inputs": [
            {"name": "asset", "type": "address"},
            {"name": "amount", "type": "uint256"},
            {"name": "interestRateMode", "type": "uint256"},
            {"name": "referralCode", "type": "uint16"},
            {"name": "onBehalfOf", "type": "address"},
        ],
        "outputs": [],
    },
    {
        "name": "repay",
        "type": "function",
        "stateMutability": "nonpayable",
        "inputs": [
            {"name": "asset", "type": "address"},
            {"name": "amount", "type": "uint256"},
            {"name": "interestRateMode", "type": "uint256"},
            {"name": "onBehalfOf", "type": "address"},
        ],
        "outputs": [{"name": "", "type": "uint256"}],
    },
    {
        "name": "setUserUseReserveAsCollateral",
        "type": "function",
        "stateMutability": "nonpayable",
        "inputs": [
            {"name": "asset", "type": "address"},
            {"name": "useAsCollateral", "type": "bool"},
        ],
        "outputs": [],
    },
]


UI_POOL_DATA_PROVIDER_ABI = [
    {
        "name": "getReservesData",
        "type": "function",
        "stateMutability": "view",
        "inputs": [{"name": "provider", "type": "address"}],
        "outputs": [
            {
                "name": "",
                "type": "tuple[]",
                "components": [
                    {"name": "underlyingAsset", "type": "address"},
                    {"name": "name", "type": "string"},
                    {"name": "symbol", "type": "string"},
                    {"name": "decimals", "type": "uint256"},
                    {"name": "baseLTVasCollateral", "type": "uint256"},
                    {"name": "reserveLiquidationThreshold", "type": "uint256"},
                    {"name": "reserveLiquidationBonus", "type": "uint256"},
                    {"name": "reserveFactor", "type": "uint256"},
                    {"name": "usageAsCollateralEnabled", "type": "bool"},
                    {"name": "borrowingEnabled", "type": "bool"},
                    {"name": "isActive", "type": "bool"},
                    {"name": "isFrozen", "type": "bool"},
                    {"name": "liquidityIndex", "type": "uint128"},
                    {"name": "variableBorrowIndex", "type": "uint128"},
                    {"name": "liquidityRate", "type": "uint128"},
                    {"name": "variableBorrowRate", "type": "uint128"},
                    {"name": "lastUpdateTimestamp", "type": "uint40"},
                    {"name": "aTokenAddress", "type": "address"},
                    {"name": "variableDebtTokenAddress", "type": "address"},
                    {"name": "interestRateStrategyAddress", "type": "address"},
                    {"name": "availableLiquidity", "type": "uint256"},
                    {"name": "totalScaledVariableDebt", "type": "uint256"},
                    {"name": "priceInMarketReferenceCurrency", "type": "uint256"},
                    {"name": "priceOracle", "type": "address"},
                    {"name": "variableRateSlope1", "type": "uint256"},
                    {"name": "variableRateSlope2", "type": "uint256"},
                    {"name": "baseVariableBorrowRate", "type": "uint256"},
                    {"name": "optimalUsageRatio", "type": "uint256"},
                    {"name": "isPaused", "type": "bool"},
                    {"name": "isSiloedBorrowing", "type": "bool"},
                    {"name": "accruedToTreasury", "type": "uint128"},
                    {"name": "unbacked", "type": "uint128"},
                    {"name": "flashLoanEnabled", "type": "bool"},
                    {"name": "debtCeiling", "type": "uint256"},
                    {"name": "debtCeilingDecimals", "type": "uint256"},
                    {"name": "borrowCap", "type": "uint256"},
                    {"name": "supplyCap", "type": "uint256"},
                    {"name": "borrowableInIsolation", "type": "bool"},
                    {"name": "virtualUnderlyingBalance", "type": "uint128"},
                    {"name": "isolationModeTotalDebt", "type": "uint128"},
                ],
            },
            {
                "name": "",
                "type": "tuple",
                "components": [
                    {"name": "marketReferenceCurrencyUnit", "type": "uint256"},
                    {"name": "marketReferenceCurrencyPriceInUsd", "type": "int256"},
                    {"name": "networkBaseTokenPriceInUsd", "type": "int256"},
                    {"name": "networkBaseTokenPriceDecimals", "type": "uint8"},
                ],
            },
        ],
    },
    {
        "name": "getUserReservesData",
        "type": "function",
        "stateMutability": "view",
        "inputs": [
            {"name": "provider", "type": "address"},
            {"name": "user", "type": "address"},
        ],
        "outputs": [
            {
                "name": "",
                "type": "tuple[]",
                "components": [
                    {"name": "underlyingAsset", "type": "address"},
                    {"name": "scaledATokenBalance", "type": "uint256"},
                    {"name": "usageAsCollateralEnabledOnUser", "type": "bool"},
                    {"name": "scaledVariableDebt", "type": "uint256"},
                ],
            },
            {"name": "", "type": "uint8"},
        ],
    },
]

UI_POOL_RESERVE_KEYS = [
    c.get("name")
    for c in (
        UI_POOL_DATA_PROVIDER_ABI[0].get("outputs", [{}])[0].get("components") or []
    )
    if c.get("name")
]


UI_INCENTIVE_DATA_PROVIDER_V3_ABI = [
    {
        "name": "getReservesIncentivesData",
        "type": "function",
        "stateMutability": "view",
        "inputs": [{"name": "provider", "type": "address"}],
        "outputs": [
            {
                "name": "",
                "type": "tuple[]",
                "components": [
                    {"name": "underlyingAsset", "type": "address"},
                    {
                        "name": "aIncentiveData",
                        "type": "tuple",
                        "components": [
                            {"name": "tokenAddress", "type": "address"},
                            {"name": "incentiveControllerAddress", "type": "address"},
                            {
                                "name": "rewardsTokenInformation",
                                "type": "tuple[]",
                                "components": [
                                    {"name": "rewardTokenSymbol", "type": "string"},
                                    {
                                        "name": "rewardTokenAddress",
                                        "type": "address",
                                    },
                                    {"name": "rewardOracleAddress", "type": "address"},
                                    {"name": "emissionPerSecond", "type": "uint256"},
                                    {
                                        "name": "incentivesLastUpdateTimestamp",
                                        "type": "uint256",
                                    },
                                    {
                                        "name": "tokenIncentivesIndex",
                                        "type": "uint256",
                                    },
                                    {"name": "emissionEndTimestamp", "type": "uint256"},
                                    {"name": "rewardPriceFeed", "type": "int256"},
                                    {"name": "rewardTokenDecimals", "type": "uint8"},
                                    {"name": "precision", "type": "uint8"},
                                    {"name": "priceFeedDecimals", "type": "uint8"},
                                ],
                            },
                        ],
                    },
                    {
                        "name": "vIncentiveData",
                        "type": "tuple",
                        "components": [
                            {"name": "tokenAddress", "type": "address"},
                            {"name": "incentiveControllerAddress", "type": "address"},
                            {
                                "name": "rewardsTokenInformation",
                                "type": "tuple[]",
                                "components": [
                                    {"name": "rewardTokenSymbol", "type": "string"},
                                    {
                                        "name": "rewardTokenAddress",
                                        "type": "address",
                                    },
                                    {"name": "rewardOracleAddress", "type": "address"},
                                    {"name": "emissionPerSecond", "type": "uint256"},
                                    {
                                        "name": "incentivesLastUpdateTimestamp",
                                        "type": "uint256",
                                    },
                                    {
                                        "name": "tokenIncentivesIndex",
                                        "type": "uint256",
                                    },
                                    {"name": "emissionEndTimestamp", "type": "uint256"},
                                    {"name": "rewardPriceFeed", "type": "int256"},
                                    {"name": "rewardTokenDecimals", "type": "uint8"},
                                    {"name": "precision", "type": "uint8"},
                                    {"name": "priceFeedDecimals", "type": "uint8"},
                                ],
                            },
                        ],
                    },
                ],
            }
        ],
    },
    {
        "name": "getUserReservesIncentivesData",
        "type": "function",
        "stateMutability": "view",
        "inputs": [
            {"name": "provider", "type": "address"},
            {"name": "user", "type": "address"},
        ],
        "outputs": [
            {
                "name": "",
                "type": "tuple[]",
                "components": [
                    {"name": "underlyingAsset", "type": "address"},
                    {
                        "name": "aTokenIncentivesUserData",
                        "type": "tuple",
                        "components": [
                            {"name": "tokenAddress", "type": "address"},
                            {"name": "incentiveControllerAddress", "type": "address"},
                            {
                                "name": "userRewardsInformation",
                                "type": "tuple[]",
                                "components": [
                                    {"name": "rewardTokenSymbol", "type": "string"},
                                    {"name": "rewardOracleAddress", "type": "address"},
                                    {
                                        "name": "rewardTokenAddress",
                                        "type": "address",
                                    },
                                    {
                                        "name": "userUnclaimedRewards",
                                        "type": "uint256",
                                    },
                                    {
                                        "name": "tokenIncentivesUserIndex",
                                        "type": "uint256",
                                    },
                                    {"name": "rewardPriceFeed", "type": "int256"},
                                    {"name": "priceFeedDecimals", "type": "uint8"},
                                    {"name": "rewardTokenDecimals", "type": "uint8"},
                                ],
                            },
                        ],
                    },
                    {
                        "name": "vTokenIncentivesUserData",
                        "type": "tuple",
                        "components": [
                            {"name": "tokenAddress", "type": "address"},
                            {"name": "incentiveControllerAddress", "type": "address"},
                            {
                                "name": "userRewardsInformation",
                                "type": "tuple[]",
                                "components": [
                                    {"name": "rewardTokenSymbol", "type": "string"},
                                    {"name": "rewardOracleAddress", "type": "address"},
                                    {
                                        "name": "rewardTokenAddress",
                                        "type": "address",
                                    },
                                    {
                                        "name": "userUnclaimedRewards",
                                        "type": "uint256",
                                    },
                                    {
                                        "name": "tokenIncentivesUserIndex",
                                        "type": "uint256",
                                    },
                                    {"name": "rewardPriceFeed", "type": "int256"},
                                    {"name": "priceFeedDecimals", "type": "uint8"},
                                    {"name": "rewardTokenDecimals", "type": "uint8"},
                                ],
                            },
                        ],
                    },
                ],
            }
        ],
    },
]


REWARDS_CONTROLLER_ABI = [
    {
        "name": "claimAllRewards",
        "type": "function",
        "stateMutability": "nonpayable",
        "inputs": [
            {"name": "assets", "type": "address[]"},
            {"name": "to", "type": "address"},
        ],
        "outputs": [
            {"name": "rewardsList", "type": "address[]"},
            {"name": "claimedAmounts", "type": "uint256[]"},
        ],
    },
    {
        "name": "claimRewards",
        "type": "function",
        "stateMutability": "nonpayable",
        "inputs": [
            {"name": "assets", "type": "address[]"},
            {"name": "amount", "type": "uint256"},
            {"name": "to", "type": "address"},
            {"name": "reward", "type": "address"},
        ],
        "outputs": [{"name": "", "type": "uint256"}],
    },
]


WRAPPED_TOKEN_GATEWAY_V3_ABI = [
    {
        "type": "function",
        "stateMutability": "view",
        "name": "getWETHAddress",
        "inputs": [],
        "outputs": [{"type": "address"}],
    },
    {
        "type": "function",
        "stateMutability": "payable",
        "name": "depositETH",
        "inputs": [
            {"name": "pool", "type": "address"},
            {"name": "onBehalfOf", "type": "address"},
            {"name": "referralCode", "type": "uint16"},
        ],
        "outputs": [],
    },
    {
        "type": "function",
        "stateMutability": "nonpayable",
        "name": "withdrawETH",
        "inputs": [
            {"name": "pool", "type": "address"},
            {"name": "amount", "type": "uint256"},
            {"name": "to", "type": "address"},
        ],
        "outputs": [],
    },
    {
        "type": "function",
        "stateMutability": "payable",
        "name": "repayETH",
        "inputs": [
            {"name": "pool", "type": "address"},
            {"name": "amount", "type": "uint256"},
            {"name": "rateMode", "type": "uint256"},
            {"name": "onBehalfOf", "type": "address"},
        ],
        "outputs": [],
    },
    {
        "type": "function",
        "stateMutability": "nonpayable",
        "name": "borrowETH",
        "inputs": [
            {"name": "pool", "type": "address"},
            {"name": "amount", "type": "uint256"},
            {"name": "interestRateMode", "type": "uint256"},
            {"name": "referralCode", "type": "uint16"},
        ],
        "outputs": [],
    },
]


WETH_ABI = [
    {
        "type": "function",
        "stateMutability": "payable",
        "name": "deposit",
        "inputs": [],
        "outputs": [],
    },
    {
        "type": "function",
        "stateMutability": "nonpayable",
        "name": "withdraw",
        "inputs": [{"name": "amount", "type": "uint256"}],
        "outputs": [],
    },
]


CHAINLINK_AGGREGATOR_ABI = [
    {
        "type": "function",
        "stateMutability": "view",
        "name": "decimals",
        "inputs": [],
        "outputs": [{"type": "uint8"}],
    },
    {
        "type": "function",
        "stateMutability": "view",
        "name": "latestAnswer",
        "inputs": [],
        "outputs": [{"type": "int256"}],
    },
]



from __future__ import annotations

import re
from typing import Any, cast

from wayfinder_paths.core.clients.TokenClient import TOKEN_CLIENT
from wayfinder_paths.core.constants import ZERO_ADDRESS
from wayfinder_paths.core.constants.base import NATIVE_COINGECKO_IDS, NATIVE_GAS_SYMBOLS
from wayfinder_paths.core.constants.chains import CHAIN_CODE_TO_ID, CHAIN_ID_TO_CODE
from wayfinder_paths.core.utils.token_refs import (
    looks_like_evm_address,
    parse_token_id_to_chain_and_address,
)
from wayfinder_paths.core.utils.tokens import get_token_decimals, is_native_token

_SIMPLE_CHAIN_SUFFIX_RE = re.compile(r"^[a-z0-9]+\s+[a-z0-9-]+$", re.IGNORECASE)


def _normalize_token_query(query: str) -> str:
    q = " ".join(str(query).strip().split())
    if not q or "-" in q or "_" in q:
        return q
    if not _SIMPLE_CHAIN_SUFFIX_RE.match(q):
        return q
    asset, chain_code = q.rsplit(" ", 1)
    if chain_code.lower() in CHAIN_CODE_TO_ID:
        return f"{asset}-{chain_code}"
    return q


def _is_gas_token(meta: dict[str, Any]) -> bool:
    asset_id = str(meta.get("asset_id") or "").lower()
    symbol = str(meta.get("symbol") or "").lower()
    return asset_id in NATIVE_COINGECKO_IDS or symbol in NATIVE_GAS_SYMBOLS


def _split_asset_chain(query: str) -> tuple[str, str] | None:
    q = str(query).strip().lower()
    if not q:
        return None

    normalized = q.replace("_", "-")
    parts = normalized.split("-")
    if len(parts) < 2:
        return None
    if parts[-1] in CHAIN_CODE_TO_ID:
        return "-".join(parts[:-1]), parts[-1]
    if parts[0] in CHAIN_CODE_TO_ID:
        return "-".join(parts[1:]), parts[0]
    return None


def _chain_id_from_meta(meta: dict[str, Any]) -> int | None:
    if meta.get("chain_id") is not None:
        try:
            return int(meta.get("chain_id"))
        except (TypeError, ValueError):
            return None

    chain = meta.get("chain") or {}
    if not isinstance(chain, dict):
        return None

    for key in ("chain_id", "chainId", "id"):
        if chain.get(key) is None:
            continue
        try:
            return int(chain.get(key))
        except (TypeError, ValueError):
            return None

    return None


def _infer_chain_code_from_query(query: str, meta: dict[str, Any]) -> str | None:
    q = str(query).strip().lower()
    if not q:
        return None

    candidates: set[str] = {str(k).lower() for k in CHAIN_CODE_TO_ID.keys()}

    addrs = meta.get("addresses") or {}
    if isinstance(addrs, dict):
        candidates.update(str(k).lower() for k in addrs.keys())

    chain_addrs = meta.get("chain_addresses") or {}
    if isinstance(chain_addrs, dict):
        candidates.update(str(k).lower() for k in chain_addrs.keys())

    best: str | None = None
    for code in candidates:
        if q.endswith(f"-{code}"):
            if best is None or len(code) > len(best):
                best = code
    return best


def _address_for_chain(meta: dict[str, Any], chain_code: str) -> str | None:
    addrs = meta.get("addresses") or {}
    if isinstance(addrs, dict):
        for key, val in addrs.items():
            if str(key).lower() == chain_code and val:
                return str(val)

    chain_addrs = meta.get("chain_addresses") or {}
    if isinstance(chain_addrs, dict):
        for key, val in chain_addrs.items():
            if str(key).lower() != chain_code:
                continue
            if isinstance(val, dict):
                addr = val.get("address")
                if addr:
                    return str(addr)

    return None


def _normalize_token_address(token_address: str | None) -> str | None:
    if not token_address:
        return None
    addr = str(token_address).strip()
    if not addr:
        return None
    return ZERO_ADDRESS if is_native_token(addr) else addr


def _select_chain_and_address(
    meta: dict[str, Any], *, query: str
) -> tuple[int | None, str | None]:
    chain_id = _chain_id_from_meta(meta)
    token_address = meta.get("address")

    desired_chain = _infer_chain_code_from_query(query, meta)
    if desired_chain:
        addr = _address_for_chain(meta, desired_chain)
        if addr:
            token_address = addr
        if desired_chain in CHAIN_CODE_TO_ID:
            chain_id = CHAIN_CODE_TO_ID[desired_chain]

    token_address_out = _normalize_token_address(
        str(token_address).strip() if token_address else None
    )
    if token_address_out is None and _is_gas_token(meta):
        token_address_out = ZERO_ADDRESS

    return chain_id, token_address_out


class TokenResolver:
    _token_details_cache: dict[str, dict[str, Any]] = {}
    _gas_token_cache: dict[str, dict[str, Any]] = {}

    @classmethod
    def _token_cache_key(cls, query: str, chain_id: int | None) -> str:
        return f"{int(chain_id)}:{query}" if chain_id is not None else query

    @classmethod
    async def _get_token_details_cached(
        cls, query: str, *, chain_id: int | None
    ) -> dict[str, Any]:
        key = cls._token_cache_key(query, chain_id)
        cached = cls._token_details_cache.get(key)
        if cached:
            return cached

        meta = await TOKEN_CLIENT.get_token_details(query, chain_id=chain_id)
        meta_out = cast(dict[str, Any], meta)
        cls._token_details_cache[key] = meta_out
        return meta_out

    @classmethod
    async def _get_gas_token_cached(cls, chain_code: str) -> dict[str, Any]:
        key = str(chain_code).strip().lower()
        cached = cls._gas_token_cache.get(key)
        if cached:
            return cached
        meta = await TOKEN_CLIENT.get_gas_token(key)
        meta_out = cast(dict[str, Any], meta)
        cls._gas_token_cache[key] = meta_out
        return meta_out

    @staticmethod
    def _validate_chain_id_hint(chain_id: int | None, *, query: str) -> int:
        if chain_id is None:
            raise ValueError(f"chain_id is required to resolve: {query}")
        chain_id_i = int(chain_id)
        if chain_id_i <= 0:
            raise ValueError(f"chain_id is required to resolve: {query}")
        return chain_id_i

    @classmethod
    async def resolve_token(
        cls, query: str, *, chain_id: int | None = None
    ) -> tuple[int, str]:
        q_raw = str(query or "").strip()
        if not q_raw:
            raise ValueError("token query is required")

        if q_raw.lower() == "native":
            return cls._validate_chain_id_hint(chain_id, query=q_raw), ZERO_ADDRESS

        parsed_chain_id, parsed_address = parse_token_id_to_chain_and_address(q_raw)
        if parsed_chain_id is not None and parsed_address is not None:
            addr = _normalize_token_address(parsed_address)
            if not addr:
                raise ValueError(f"Cannot resolve token: {query}")
            return int(parsed_chain_id), addr

        if looks_like_evm_address(q_raw):
            chain_id_i = cls._validate_chain_id_hint(chain_id, query=q_raw)
            addr = _normalize_token_address(q_raw)
            if not addr:
                raise ValueError(f"Cannot resolve token: {query}")
            return chain_id_i, addr

        q = _normalize_token_query(q_raw)
        meta: dict[str, Any] | None = None

        split = _split_asset_chain(q)
        if split:
            asset, chain_code = split
            if asset in NATIVE_COINGECKO_IDS or asset in NATIVE_GAS_SYMBOLS:
                try:
                    gas_meta = await cls._get_gas_token_cached(chain_code)
                    if isinstance(gas_meta, dict) and _is_gas_token(gas_meta):
                        meta = gas_meta
                except Exception:
                    meta = None

        if meta is None:
            meta = await cls._get_token_details_cached(q, chain_id=chain_id)

        chain_out, addr_out = _select_chain_and_address(meta, query=q)
        if chain_out is None or not addr_out:
            raise ValueError(f"Cannot resolve token: {query}")
        return int(chain_out), str(addr_out)

    @classmethod
    async def resolve_token_meta(
        cls, query: str, *, chain_id: int | None = None
    ) -> dict[str, Any]:
        q_raw = str(query or "").strip()
        if not q_raw:
            raise ValueError("token query is required")

        meta: dict[str, Any] | None = None
        chain_code: str = None
        if q_raw.lower() == "native":
            if chain_id:
                chain_id_i = cls._validate_chain_id_hint(chain_id, query=q_raw)
                chain_code = CHAIN_ID_TO_CODE.get(chain_id_i)
                gas_meta = await cls._get_gas_token_cached(chain_code)
                if isinstance(gas_meta, dict) and _is_gas_token(gas_meta):
                    meta = gas_meta
            else:
                raise ValueError("Chain id is not provided for native query")

        parsed_chain_id, parsed_address = parse_token_id_to_chain_and_address(q_raw)
        if parsed_chain_id is not None and parsed_address is not None:
            addr = _normalize_token_address(parsed_address)
            if not addr:
                raise ValueError(f"Cannot resolve token: {query}")
            decimals = await get_token_decimals(addr, int(parsed_chain_id))
            return {
                "token_id": q_raw,
                "symbol": q_raw,
                "decimals": int(decimals),
                "chain_id": int(parsed_chain_id),
                "address": addr,
                "metadata": {"source": "local"},
            }

        if looks_like_evm_address(q_raw):
            chain_id_i = cls._validate_chain_id_hint(chain_id, query=q_raw)
            addr = _normalize_token_address(q_raw)
            if not addr:
                raise ValueError(f"Cannot resolve token: {query}")
            decimals = await get_token_decimals(addr, int(chain_id_i))
            return {
                "token_id": q_raw,
                "symbol": q_raw,
                "decimals": int(decimals),
                "chain_id": int(chain_id_i),
                "address": addr,
                "metadata": {"source": "address"},
            }

        q = _normalize_token_query(q_raw)

        split = _split_asset_chain(q)
        if split:
            asset, chain_code = split
            if chain_code.lower() == "native":
                chain_code = asset
                asset = "native"
            if (
                asset in NATIVE_COINGECKO_IDS
                or asset in NATIVE_GAS_SYMBOLS
                or asset == "native"
            ):
                try:
                    gas_meta = await cls._get_gas_token_cached(chain_code)
                    if isinstance(gas_meta, dict) and _is_gas_token(gas_meta):
                        meta = gas_meta
                except Exception:
                    meta = None
            else:
                try:
                    meta = await cls._get_token_details_cached(
                        asset, chain_id=CHAIN_CODE_TO_ID[chain_code]
                    )
                except Exception as e:
                    raise ValueError(
                        f"Cannot resolve token: {query}, chain_id: {chain_id}"
                    ) from e

        if meta is None:
            try:
                if chain_id is None and chain_code is not None:
                    chain_id = CHAIN_CODE_TO_ID[chain_code]
                meta = await cls._get_token_details_cached(q, chain_id=chain_id)
            except Exception as e:
                raise ValueError(
                    f"Cannot resolve token: {query}, chain_id: {chain_id}"
                ) from e

        chain_out, addr_out = _select_chain_and_address(meta, query=q)
        if chain_out is None or not addr_out:
            raise ValueError(f"Cannot resolve token: {query}")

        meta_out = dict(meta)
        meta_out["chain_id"] = int(chain_out)
        meta_out["address"] = str(addr_out)
        meta_out.setdefault("metadata", {})
        if isinstance(meta_out["metadata"], dict):
            meta_out["metadata"].setdefault("source", "api")  # type: ignore[union-attr]
            meta_out["metadata"].setdefault("query_normalized", q)  # type: ignore[union-attr]
        return meta_out



from __future__ import annotations

from unittest.mock import AsyncMock, patch

import pytest

from wayfinder_paths.mcp.tools.quotes import quote_swap


@pytest.mark.asyncio
async def test_quote_swap_returns_compact_best_quote_by_default():
    fake_wallet = {"address": "0x000000000000000000000000000000000000dEaD"}

    from_meta = {
        "token_id": "ethereum-arbitrum",
        "asset_id": "ethereum",
        "symbol": "ETH",
        "decimals": 18,
        "chain_id": 42161,
        "address": "0x0000000000000000000000000000000000000000",
    }
    to_meta = {
        "token_id": "usd-coin-arbitrum",
        "asset_id": "usd-coin",
        "symbol": "USDC",
        "decimals": 6,
        "chain_id": 42161,
        "address": "0xaf88d065e77c8cc2239327c5edb3a432268e5831",
    }

    async def fake_resolve(query: str, *, chain_id: int | None = None):
        _ = chain_id
        if "ethereum" in query.lower() or "eth" in query.lower():
            return from_meta
        return to_meta

    fake_brap = AsyncMock()
    calldata = {"data": "0x" + ("ab" * 4096)}
    fake_brap.get_quote = AsyncMock(
        return_value={
            "quotes": {
                "quote_count": 3,
                "best_quote": {
                    "provider": "brap_best",
                    "input_amount": "1700000000000000",
                    "output_amount": "1234567",
                    "input_amount_usd": 5.0,
                    "output_amount_usd": 4.99,
                    "gas_estimate": 210000,
                    "fee_estimate": {"total_usd": 0.01},
                    "native_input": True,
                    "native_output": False,
                    "calldata": calldata,
                    "wrap_transaction": None,
                    "unwrap_transaction": None,
                },
                "all_quotes": [
                    {"provider": "brap_best"},
                    {"provider": "brap_alt"},
                    {"provider": "brap_alt"},
                ],
            }
        }
    )

    with (
        patch(
            "wayfinder_paths.mcp.tools.quotes.find_wallet_by_label",
            return_value=fake_wallet,
        ),
        patch(
            "wayfinder_paths.mcp.tools.quotes.TokenResolver.resolve_token_meta",
            new_callable=AsyncMock,
            side_effect=fake_resolve,
        ),
        patch("wayfinder_paths.mcp.tools.quotes.BRAP_CLIENT", fake_brap),
    ):
        out = await quote_swap(
            wallet_label="main",
            from_token="ethereum-arbitrum",
            to_token="usd-coin-arbitrum",
            amount="0.0017",
            slippage_bps=50,
        )

    assert out["ok"] is True
    res = out["result"]
    assert "raw" not in res["quote"]

    best = res["quote"]["best_quote"]
    assert best["provider"] == "brap_best"
    assert best["output_amount"] == "1234567"
    assert best["calldata_len"] > 0
    assert "calldata" not in best
    assert res["quote"]["quote_count"] == 3
    assert res["quote"]["providers"] == ["brap_best", "brap_alt"]


@pytest.mark.asyncio
async def test_quote_swap_can_include_calldata_when_requested():
    fake_wallet = {"address": "0x000000000000000000000000000000000000dEaD"}

    from_meta = {
        "token_id": "ethereum-arbitrum",
        "asset_id": "ethereum",
        "symbol": "ETH",
        "decimals": 18,
        "chain_id": 42161,
        "address": "0x0000000000000000000000000000000000000000",
    }
    to_meta = {
        "token_id": "usd-coin-arbitrum",
        "asset_id": "usd-coin",
        "symbol": "USDC",
        "decimals": 6,
        "chain_id": 42161,
        "address": "0xaf88d065e77c8cc2239327c5edb3a432268e5831",
    }

    async def fake_resolve(query: str, *, chain_id: int | None = None):
        _ = chain_id
        if "ethereum" in query.lower() or "eth" in query.lower():
            return from_meta
        return to_meta

    calldata = {"data": "0x" + ("cd" * 1024)}
    fake_brap = AsyncMock()
    fake_brap.get_quote = AsyncMock(
        return_value={
            "quotes": {
                "quote_count": 1,
                "best_quote": {
                    "provider": "brap_best",
                    "output_amount": "1",
                    "calldata": calldata,
                },
                "all_quotes": [{"provider": "brap_best"}],
            }
        }
    )

    with (
        patch(
            "wayfinder_paths.mcp.tools.quotes.find_wallet_by_label",
            return_value=fake_wallet,
        ),
        patch(
            "wayfinder_paths.mcp.tools.quotes.TokenResolver.resolve_token_meta",
            new_callable=AsyncMock,
            side_effect=fake_resolve,
        ),
        patch("wayfinder_paths.mcp.tools.quotes.BRAP_CLIENT", fake_brap),
    ):
        out = await quote_swap(
            wallet_label="main",
            from_token="ethereum-arbitrum",
            to_token="usd-coin-arbitrum",
            amount="0.0017",
            slippage_bps=50,
            include_calldata=True,
        )

    assert out["ok"] is True
    best = out["result"]["quote"]["best_quote"]
    assert best["calldata"] == calldata["data"]


@pytest.mark.asyncio
async def test_quote_swap_accepts_top_level_brap_shape():
    fake_wallet = {"address": "0x000000000000000000000000000000000000dEaD"}

    from_meta = {
        "token_id": "usd-coin-arbitrum",
        "asset_id": "usd-coin",
        "symbol": "USDC",
        "decimals": 6,
        "chain_id": 42161,
        "address": "0xaf88d065e77c8cc2239327c5edb3a432268e5831",
    }
    to_meta = {
        "token_id": "tether-arbitrum",
        "asset_id": "tether",
        "symbol": "USDT",
        "decimals": 6,
        "chain_id": 42161,
        "address": "0xFd086bC7CD5C481DCC9C85ebE478A1C0b69FCbb9",
    }

    async def fake_resolve(query: str, *, chain_id: int | None = None):
        _ = chain_id
        if "tether" in query.lower() or "usdt" in query.lower():
            return to_meta
        return from_meta

    fake_brap = AsyncMock()
    fake_brap.get_quote = AsyncMock(
        return_value={
            "quotes": [{"provider": "brap_best"}, {"provider": "brap_alt"}],
            "best_quote": {
                "provider": "brap_best",
                "output_amount": "1",
                "calldata": {"data": "0xabc"},
            },
        }
    )

    with (
        patch(
            "wayfinder_paths.mcp.tools.quotes.find_wallet_by_label",
            return_value=fake_wallet,
        ),
        patch(
            "wayfinder_paths.mcp.tools.quotes.TokenResolver.resolve_token_meta",
            new_callable=AsyncMock,
            side_effect=fake_resolve,
        ),
        patch("wayfinder_paths.mcp.tools.quotes.BRAP_CLIENT", fake_brap),
    ):
        out = await quote_swap(
            wallet_label="main",
            from_token="usd-coin-arbitrum",
            to_token="tether-arbitrum",
            amount="1",
        )

    assert out["ok"] is True
    assert out["result"]["quote"]["quote_count"] == 2
    assert out["result"]["quote"]["providers"] == ["brap_best", "brap_alt"]



from wayfinder_paths.core.utils.evm_helpers import get_abi_filtered


async def allow_functions(
    policy_name: str,
    abi_chain_id: int,
    address: str,
    function_names: list[str],
    abi_address_override: str = None,
    manual_abi: dict = None,
):
    # Note: ChainID is just for fetching ABI, doesn't appear in the final policy. Doesn't bind a strict chain.
    return {
        "name": policy_name,
        "method": "eth_signTransaction",
        "action": "ALLOW",
        "conditions": [
            {
                "field_source": "ethereum_transaction",
                "field": "to",
                "operator": "eq",
                "value": address,
            },
            {
                "field_source": "ethereum_calldata",
                "field": "function_name",
                "abi": manual_abi
                or await get_abi_filtered(
                    abi_chain_id, abi_address_override or address, function_names
                ),
                "operator": "in",
                "value": function_names,
            },
        ],
    }



from __future__ import annotations

from typing import Any

from eth_utils import to_checksum_address

from wayfinder_paths.core.adapters.BaseAdapter import BaseAdapter
from wayfinder_paths.core.constants.aave_v3_abi import (
    POOL_ABI,
    REWARDS_CONTROLLER_ABI,
    UI_INCENTIVE_DATA_PROVIDER_V3_ABI,
    UI_POOL_DATA_PROVIDER_ABI,
    UI_POOL_RESERVE_KEYS,
    WETH_ABI,
    WRAPPED_TOKEN_GATEWAY_V3_ABI,
)
from wayfinder_paths.core.constants.aave_v3_contracts import AAVE_V3_BY_CHAIN
from wayfinder_paths.core.constants.base import MAX_UINT256, SECONDS_PER_YEAR
from wayfinder_paths.core.utils import web3 as web3_utils
from wayfinder_paths.core.utils.interest import RAY, apr_to_apy, ray_to_apr
from wayfinder_paths.core.utils.symbols import is_stable_symbol, normalize_symbol
from wayfinder_paths.core.utils.tokens import ensure_allowance, get_token_balance
from wayfinder_paths.core.utils.transaction import encode_call, send_transaction

VARIABLE_RATE_MODE = 2
REFERRAL_CODE = 0


def _reserve_to_dict(reserve: Any, reserve_keys: list[str]) -> dict[str, Any]:
    if isinstance(reserve, dict):
        return dict(reserve)
    return dict(zip(reserve_keys, reserve, strict=False))


def _compute_supply_cap_headroom(
    reserve: dict[str, Any], decimals: int
) -> tuple[int | None, int | None]:
    supply_cap_tokens = int(reserve.get("supplyCap") or 0)
    if supply_cap_tokens <= 0:
        return (None, None)
    unit = 10 ** max(0, int(decimals))
    supply_cap_wei = supply_cap_tokens * unit

    available = int(reserve.get("availableLiquidity") or 0)
    scaled_variable_debt = int(reserve.get("totalScaledVariableDebt") or 0)
    variable_index = int(reserve.get("variableBorrowIndex") or 0)
    current_variable_debt = (scaled_variable_debt * variable_index) // RAY

    total_supplied = available + current_variable_debt
    headroom = supply_cap_wei - total_supplied
    if headroom < 0:
        headroom = 0
    return (headroom, supply_cap_tokens)


def _base_currency_to_ref(base_currency: Any) -> tuple[int, float]:
    # base_currency: (marketReferenceCurrencyUnit, marketReferenceCurrencyPriceInUsd, networkBaseTokenPriceInUsd, networkBaseTokenPriceDecimals)
    try:
        ref_unit = int(base_currency[0]) if base_currency else 1
    except (TypeError, ValueError):
        ref_unit = 1
    if not ref_unit:
        ref_unit = 1

    try:
        ref_usd_raw = int(base_currency[1]) if base_currency else 0
    except (TypeError, ValueError):
        ref_usd_raw = 0

    try:
        ref_usd_decimals = int(base_currency[3]) if base_currency else 0
    except (TypeError, ValueError):
        ref_usd_decimals = 0

    ref_usd = (
        ref_usd_raw / (10**ref_usd_decimals)
        if ref_usd_decimals and ref_usd_decimals > 0
        else float(ref_usd_raw)
    )
    return (ref_unit, float(ref_usd))


def _reward_rows(rewards_info: Any) -> list[Any]:
    try:
        return list(rewards_info[2] or [])
    except Exception:
        return []


def _compute_incentive_apr(
    rewards_info: Any,
    *,
    denom: float,
    side: str,
    include_details: bool,
) -> tuple[float, list[dict[str, Any]]]:
    apr_total = 0.0
    incentives: list[dict[str, Any]] = []

    for reward in _reward_rows(rewards_info):
        try:
            reward_addr = to_checksum_address(str(reward[1]))
            emission_per_second = int(reward[3] or 0)
            end_ts = int(reward[6] or 0)
            reward_price_feed = int(reward[7] or 0)
            reward_token_decimals = int(reward[8] or 0)
            price_feed_decimals = int(reward[10] or 0)
        except Exception:  # noqa: BLE001
            continue

        price_usd_r = (
            float(reward_price_feed) / (10**price_feed_decimals)
            if price_feed_decimals and reward_price_feed
            else 0.0
        )

        annual_rewards_usd = (
            (float(emission_per_second) / (10**reward_token_decimals))
            * SECONDS_PER_YEAR
            * price_usd_r
            if emission_per_second and reward_token_decimals >= 0 and price_usd_r
            else 0.0
        )
        apr = (
            float(annual_rewards_usd) / float(denom)
            if denom and annual_rewards_usd
            else 0.0
        )
        if apr:
            apr_total += apr

        if include_details:
            incentives.append(
                {
                    "side": side,
                    "token": reward_addr,
                    "symbol": str(reward[0] or ""),
                    "apr": float(apr),
                    "emission_per_second": int(emission_per_second),
                    "distribution_end": int(end_ts) or None,
                    "price_usd": float(price_usd_r),
                }
            )

    return float(apr_total), incentives


class AaveV3Adapter(BaseAdapter):
    adapter_type = "AAVE_V3"

    def __init__(
        self,
        config: dict[str, Any] | None = None,
        strategy_wallet_signing_callback=None,
    ) -> None:
        super().__init__("aave_v3_adapter", config or {})
        cfg = config or {}
        self.strategy_wallet_signing_callback = strategy_wallet_signing_callback

        strategy_addr = (cfg.get("strategy_wallet") or {}).get("address")
        self.strategy_wallet_address: str | None = (
            to_checksum_address(strategy_addr) if strategy_addr else None
        )

        # Cache: (chain_id, underlying.lower()) -> variableDebtTokenAddress
        self._variable_debt_token_by_chain_underlying: dict[tuple[int, str], str] = {}
        # Cache: chain_id -> wrapped native token address (WETH/WBNB/WMATIC/etc)
        self._wrapped_native_by_chain: dict[int, str] = {}

    def _entry(self, chain_id: int) -> dict[str, str]:
        entry = AAVE_V3_BY_CHAIN.get(int(chain_id))
        if not entry:
            raise ValueError(f"Unsupported Aave v3 chain_id={chain_id}")
        return entry

    async def _pool(self, *, chain_id: int) -> str:
        return to_checksum_address(self._entry(int(chain_id))["pool"])

    async def _provider(self, *, chain_id: int) -> str:
        return to_checksum_address(
            self._entry(int(chain_id))["pool_addresses_provider"]
        )

    async def _ui_pool(self, *, chain_id: int) -> str:
        return to_checksum_address(self._entry(int(chain_id))["ui_pool_data_provider"])

    async def _ui_incentives(self, *, chain_id: int) -> str:
        return to_checksum_address(
            self._entry(int(chain_id))["ui_incentive_data_provider"]
        )

    async def _rewards_controller(self, *, chain_id: int) -> str:
        return to_checksum_address(self._entry(int(chain_id))["rewards_controller"])

    async def _wrapped_native(self, *, chain_id: int) -> str:
        cached = self._wrapped_native_by_chain.get(int(chain_id))
        if cached:
            return cached

        gateway = self._entry(int(chain_id)).get("wrapped_token_gateway")
        if not gateway:
            raise ValueError(
                f"wrapped_token_gateway not configured for chain_id={chain_id}"
            )

        async with web3_utils.web3_from_chain_id(int(chain_id)) as web3:
            gw = web3.eth.contract(
                address=to_checksum_address(gateway),
                abi=WRAPPED_TOKEN_GATEWAY_V3_ABI,
            )
            wrapped = await gw.functions.getWETHAddress().call(
                block_identifier="pending"
            )
            wrapped = to_checksum_address(str(wrapped))
            self._wrapped_native_by_chain[int(chain_id)] = wrapped
            return wrapped

    async def get_all_markets(
        self,
        *,
        chain_id: int,
        include_rewards: bool = True,
    ) -> tuple[bool, list[dict[str, Any]] | str]:
        try:
            entry = self._entry(int(chain_id))
            ui_pool_addr = to_checksum_address(entry["ui_pool_data_provider"])
            provider_addr = to_checksum_address(entry["pool_addresses_provider"])

            reserves_incentives: dict[str, Any] = {}
            async with web3_utils.web3_from_chain_id(int(chain_id)) as web3:
                if include_rewards:
                    try:
                        ui_incentives_addr = to_checksum_address(
                            entry["ui_incentive_data_provider"]
                        )
                        ui_incentives = web3.eth.contract(
                            address=ui_incentives_addr,
                            abi=UI_INCENTIVE_DATA_PROVIDER_V3_ABI,
                        )
                        inc_rows = (
                            await ui_incentives.functions.getReservesIncentivesData(
                                provider_addr
                            ).call(block_identifier="pending")
                        )
                        for row in inc_rows or []:
                            try:
                                underlying = to_checksum_address(str(row[0]))
                            except Exception:  # noqa: BLE001
                                continue
                            reserves_incentives[underlying.lower()] = row
                    except Exception:  # noqa: BLE001
                        reserves_incentives = {}

                ui_pool = web3.eth.contract(
                    address=ui_pool_addr, abi=UI_POOL_DATA_PROVIDER_ABI
                )
                reserves, base_currency = await ui_pool.functions.getReservesData(
                    provider_addr
                ).call(block_identifier="pending")

                ref_unit, ref_usd = _base_currency_to_ref(base_currency)

                reserve_keys = UI_POOL_RESERVE_KEYS
                markets: list[dict[str, Any]] = []

                for reserve in reserves or []:
                    r = _reserve_to_dict(reserve, reserve_keys)

                    underlying = to_checksum_address(str(r.get("underlyingAsset")))
                    symbol_raw = str(r.get("symbol") or "")
                    decimals = int(r.get("decimals") or 18)
                    a_token = to_checksum_address(str(r.get("aTokenAddress")))
                    v_debt = to_checksum_address(str(r.get("variableDebtTokenAddress")))
                    self._variable_debt_token_by_chain_underlying[
                        (int(chain_id), underlying.lower())
                    ] = v_debt

                    liquidity_rate_ray = int(r.get("liquidityRate") or 0)
                    variable_borrow_rate_ray = int(r.get("variableBorrowRate") or 0)

                    price_market_ref = int(r.get("priceInMarketReferenceCurrency") or 0)
                    try:
                        price_market_ref_float = (
                            float(price_market_ref) / ref_unit if ref_unit else 0.0
                        )
                    except ZeroDivisionError:
                        price_market_ref_float = 0.0
                    price_usd = float(price_market_ref_float) * float(ref_usd)

                    supply_apr = ray_to_apr(liquidity_rate_ray)
                    borrow_apr = ray_to_apr(variable_borrow_rate_ray)

                    available_liquidity = int(r.get("availableLiquidity") or 0)
                    scaled_variable_debt = int(r.get("totalScaledVariableDebt") or 0)
                    variable_index = int(r.get("variableBorrowIndex") or 0)
                    total_variable_debt = (scaled_variable_debt * variable_index) // RAY
                    tvl = available_liquidity + total_variable_debt

                    symbol_canonical = normalize_symbol(symbol_raw) or normalize_symbol(
                        underlying
                    )

                    headroom_wei, supply_cap_tokens = _compute_supply_cap_headroom(
                        r, decimals
                    )

                    base_supply_apy = float(apr_to_apy(supply_apr))
                    base_borrow_apy = float(apr_to_apy(borrow_apr))

                    market_row: dict[str, Any] = {
                        "chain_id": int(chain_id),
                        "pool": to_checksum_address(entry["pool"]),
                        "underlying": underlying,
                        "symbol": symbol_raw,
                        "symbol_canonical": symbol_canonical,
                        "decimals": int(decimals),
                        "a_token": a_token,
                        "variable_debt_token": v_debt,
                        "is_active": bool(r.get("isActive")),
                        "is_frozen": bool(r.get("isFrozen")),
                        "is_paused": bool(r.get("isPaused")),
                        "is_siloed_borrowing": bool(r.get("isSiloedBorrowing")),
                        "is_stablecoin": bool(is_stable_symbol(symbol_raw)),
                        "usage_as_collateral_enabled": bool(
                            r.get("usageAsCollateralEnabled")
                        ),
                        "borrowing_enabled": bool(r.get("borrowingEnabled")),
                        "ltv_bps": int(r.get("baseLTVasCollateral") or 0),
                        "liquidation_threshold_bps": int(
                            r.get("reserveLiquidationThreshold") or 0
                        ),
                        "price_usd": float(price_usd),
                        "supply_apr": float(supply_apr),
                        "supply_apy": float(base_supply_apy),
                        "variable_borrow_apr": float(borrow_apr),
                        "variable_borrow_apy": float(base_borrow_apy),
                        "available_liquidity": int(available_liquidity),
                        "total_variable_debt": int(total_variable_debt),
                        "tvl": int(tvl),
                        "borrow_cap": int(r.get("borrowCap") or 0) or None,
                        "supply_cap": supply_cap_tokens,
                        "supply_cap_headroom": headroom_wei,
                    }

                    if include_rewards:
                        reward_supply_apr = 0.0
                        reward_borrow_apr = 0.0
                        incentives_out: list[dict[str, Any]] = []

                        denom_supply = (
                            (available_liquidity + total_variable_debt)
                            / (10**decimals)
                            * price_usd
                            if price_usd
                            and (available_liquidity + total_variable_debt) > 0
                            else 0.0
                        )
                        denom_borrow = (
                            total_variable_debt / (10**decimals) * price_usd
                            if price_usd and total_variable_debt > 0
                            else 0.0
                        )

                        inc_row = reserves_incentives.get(underlying.lower())
                        if inc_row:
                            a_inc = inc_row[1] if len(inc_row) > 1 else None
                            v_inc = inc_row[2] if len(inc_row) > 2 else None

                            supply_apr_inc, supply_incs = _compute_incentive_apr(
                                a_inc,
                                denom=float(denom_supply),
                                side="supply",
                                include_details=True,
                            )
                            borrow_apr_inc, borrow_incs = _compute_incentive_apr(
                                v_inc,
                                denom=float(denom_borrow),
                                side="borrow",
                                include_details=True,
                            )
                            reward_supply_apr += supply_apr_inc
                            reward_borrow_apr += borrow_apr_inc
                            incentives_out.extend(supply_incs)
                            incentives_out.extend(borrow_incs)

                        market_row["reward_supply_apr"] = float(reward_supply_apr)
                        market_row["reward_borrow_apr"] = float(reward_borrow_apr)
                        market_row["supply_apy_with_rewards"] = float(
                            base_supply_apy + reward_supply_apr
                        )
                        market_row["borrow_apy_with_rewards"] = float(
                            base_borrow_apy - reward_borrow_apr
                        )
                        market_row["incentives"] = incentives_out

                    markets.append(market_row)

                return True, markets
        except Exception as exc:  # noqa: BLE001
            return False, str(exc)

    async def get_full_user_state(
        self,
        *,
        account: str,
        include_zero_positions: bool = False,
        include_rewards: bool = True,
    ) -> tuple[bool, dict[str, Any] | str]:
        """Query all supported Aave V3 chains and merge results."""
        account = to_checksum_address(account)
        all_positions: list[dict[str, Any]] = []
        chains_queried: list[dict[str, Any]] = []
        errors: list[str] = []

        for cid in AAVE_V3_BY_CHAIN:
            ok, result = await self.get_full_user_state_per_chain(
                chain_id=cid,
                account=account,
                include_zero_positions=include_zero_positions,
                include_rewards=include_rewards,
            )
            if ok:
                chain_data = result  # type: ignore[assignment]
                all_positions.extend(chain_data.get("positions", []))
                chains_queried.append(
                    {
                        "chain_id": cid,
                        "pool": chain_data.get("pool"),
                        "userEmodeCategoryId": chain_data.get("userEmodeCategoryId", 0),
                    }
                )
            else:
                errors.append(f"chain {cid}: {result}")

        if not chains_queried and errors:
            return False, "; ".join(errors)

        return True, {
            "protocol": "aave_v3",
            "account": account,
            "chains": chains_queried,
            "positions": all_positions,
            "errors": errors,
        }

    async def get_full_user_state_per_chain(
        self,
        *,
        chain_id: int,
        account: str,
        include_zero_positions: bool = False,
        include_rewards: bool = True,
    ) -> tuple[bool, dict[str, Any] | str]:
        try:
            entry = self._entry(int(chain_id))
            ui_pool_addr = to_checksum_address(entry["ui_pool_data_provider"])
            provider_addr = to_checksum_address(entry["pool_addresses_provider"])

            account = to_checksum_address(account)

            async with web3_utils.web3_from_chain_id(int(chain_id)) as web3:
                ui_pool = web3.eth.contract(
                    address=ui_pool_addr, abi=UI_POOL_DATA_PROVIDER_ABI
                )
                reserves, base_currency = await ui_pool.functions.getReservesData(
                    provider_addr
                ).call(block_identifier="pending")
                user_reserves, user_emode = await ui_pool.functions.getUserReservesData(
                    provider_addr, account
                ).call(block_identifier="pending")

                ref_unit, ref_usd = _base_currency_to_ref(base_currency)

                reserve_keys = UI_POOL_RESERVE_KEYS
                by_underlying: dict[str, dict[str, Any]] = {}
                for reserve in reserves or []:
                    r = _reserve_to_dict(reserve, reserve_keys)
                    underlying = to_checksum_address(str(r.get("underlyingAsset")))
                    by_underlying[underlying.lower()] = r

                reserves_incentives: dict[str, Any] = {}
                user_rewards_by_underlying: dict[str, list[dict[str, Any]]] = {}
                if include_rewards:
                    try:
                        ui_incentives_addr = to_checksum_address(
                            entry["ui_incentive_data_provider"]
                        )
                        ui_incentives = web3.eth.contract(
                            address=ui_incentives_addr,
                            abi=UI_INCENTIVE_DATA_PROVIDER_V3_ABI,
                        )

                        try:
                            inc_rows = (
                                await ui_incentives.functions.getReservesIncentivesData(
                                    provider_addr
                                ).call(block_identifier="pending")
                            )
                            for row in inc_rows or []:
                                try:
                                    underlying = to_checksum_address(str(row[0]))
                                except Exception:  # noqa: BLE001
                                    continue
                                reserves_incentives[underlying.lower()] = row
                        except Exception:  # noqa: BLE001
                            reserves_incentives = {}

                        user_inc = (
                            await ui_incentives.functions.getUserReservesIncentivesData(
                                provider_addr, account
                            ).call(block_identifier="pending")
                        )

                        for row in user_inc or []:
                            try:
                                underlying = to_checksum_address(str(row[0]))
                            except Exception:  # noqa: BLE001
                                continue
                            out_rewards: list[dict[str, Any]] = []
                            for side_idx, side in (
                                (1, "supply"),
                                (2, "borrow"),
                                (3, "stable_borrow"),
                            ):
                                try:
                                    inc = row[side_idx]
                                    rewards = list(inc[2] or [])
                                except Exception:
                                    rewards = []
                                for rwd in rewards:
                                    try:
                                        out_rewards.append(
                                            {
                                                "side": side,
                                                "token": to_checksum_address(
                                                    str(rwd[2])
                                                ),
                                                "symbol": str(rwd[0] or ""),
                                                "unclaimed": int(rwd[3] or 0),
                                                "price_usd": (
                                                    float(int(rwd[5] or 0))
                                                    / (10 ** int(rwd[6] or 0))
                                                    if int(rwd[5] or 0)
                                                    and int(rwd[6] or 0)
                                                    else 0.0
                                                ),
                                                "price_feed_decimals": int(rwd[6] or 0),
                                                "reward_token_decimals": int(
                                                    rwd[7] or 0
                                                ),
                                            }
                                        )
                                    except Exception:  # noqa: BLE001
                                        continue
                            user_rewards_by_underlying[underlying.lower()] = out_rewards
                    except Exception:
                        reserves_incentives = {}
                        user_rewards_by_underlying = {}

            positions: list[dict[str, Any]] = []
            for row in user_reserves or []:
                try:
                    underlying = to_checksum_address(str(row[0]))
                except Exception:  # noqa: BLE001
                    continue

                reserve = by_underlying.get(underlying.lower())
                if not reserve:
                    continue

                decimals = int(reserve.get("decimals") or 18)
                symbol_raw = str(reserve.get("symbol") or "")

                liquidity_index = int(reserve.get("liquidityIndex") or 0)
                variable_borrow_index = int(reserve.get("variableBorrowIndex") or 0)

                scaled_supply = int(row[1] or 0)
                scaled_var_debt = int(row[3] or 0)
                stable_debt = 0
                is_collateral = bool(row[2])

                supply_raw = (
                    (scaled_supply * liquidity_index) // RAY if scaled_supply else 0
                )
                variable_debt_raw = (
                    (scaled_var_debt * variable_borrow_index) // RAY
                    if scaled_var_debt
                    else 0
                )

                price_market_ref = int(
                    reserve.get("priceInMarketReferenceCurrency") or 0
                )
                price_usd = (
                    (float(price_market_ref) / ref_unit) * float(ref_usd)
                    if ref_unit and price_market_ref
                    else 0.0
                )

                supply_usd = (
                    float(supply_raw) / (10**decimals) * price_usd
                    if supply_raw and price_usd
                    else 0.0
                )
                borrow_usd = (
                    float(variable_debt_raw) / (10**decimals) * price_usd
                    if variable_debt_raw and price_usd
                    else 0.0
                )

                if (
                    not include_zero_positions
                    and supply_raw <= 0
                    and variable_debt_raw <= 0
                    and stable_debt <= 0
                ):
                    continue

                liquidity_rate_ray = int(reserve.get("liquidityRate") or 0)
                variable_borrow_rate_ray = int(reserve.get("variableBorrowRate") or 0)
                supply_apy = float(apr_to_apy(ray_to_apr(liquidity_rate_ray)))
                variable_borrow_apy = float(
                    apr_to_apy(ray_to_apr(variable_borrow_rate_ray))
                )

                reward_supply_apr = 0.0
                reward_borrow_apr = 0.0
                if include_rewards:
                    inc_row = reserves_incentives.get(underlying.lower())
                    if inc_row:
                        a_inc = inc_row[1] if len(inc_row) > 1 else None
                        v_inc = inc_row[2] if len(inc_row) > 2 else None

                        available_liquidity = int(
                            reserve.get("availableLiquidity") or 0
                        )
                        scaled_variable_debt = int(
                            reserve.get("totalScaledVariableDebt") or 0
                        )
                        variable_index = int(reserve.get("variableBorrowIndex") or 0)
                        total_variable_debt = (
                            scaled_variable_debt * variable_index
                        ) // RAY
                        denom_supply = (
                            (available_liquidity + total_variable_debt)
                            / (10**decimals)
                            * price_usd
                            if price_usd
                            and (available_liquidity + total_variable_debt) > 0
                            else 0.0
                        )
                        denom_borrow = (
                            total_variable_debt / (10**decimals) * price_usd
                            if price_usd and total_variable_debt > 0
                            else 0.0
                        )

                        reward_supply_apr, _ = _compute_incentive_apr(
                            a_inc,
                            denom=float(denom_supply),
                            side="supply",
                            include_details=False,
                        )
                        reward_borrow_apr, _ = _compute_incentive_apr(
                            v_inc,
                            denom=float(denom_borrow),
                            side="borrow",
                            include_details=False,
                        )

                positions.append(
                    {
                        "chain_id": int(chain_id),
                        "underlying": underlying,
                        "symbol": symbol_raw,
                        "symbol_canonical": normalize_symbol(symbol_raw)
                        or normalize_symbol(underlying),
                        "decimals": decimals,
                        "a_token": to_checksum_address(
                            str(reserve.get("aTokenAddress"))
                        ),
                        "variable_debt_token": to_checksum_address(
                            str(reserve.get("variableDebtTokenAddress"))
                        ),
                        "usage_as_collateral": is_collateral,
                        "supply_raw": int(supply_raw),
                        "variable_borrow_raw": int(variable_debt_raw),
                        "stable_borrow_raw": int(stable_debt),
                        "supply_usd": float(supply_usd),
                        "variable_borrow_usd": float(borrow_usd),
                        "price_usd": float(price_usd),
                        "supply_apy": supply_apy,
                        "variable_borrow_apy": variable_borrow_apy,
                        "reward_supply_apr": reward_supply_apr,
                        "reward_borrow_apr": reward_borrow_apr,
                        "supply_apy_with_rewards": supply_apy + reward_supply_apr,
                        "borrow_apy_with_rewards": variable_borrow_apy
                        - reward_borrow_apr,
                        "rewards": user_rewards_by_underlying.get(underlying.lower())
                        or [],
                    }
                )

            return True, {
                "protocol": "aave_v3",
                "chain_id": int(chain_id),
                "pool": entry["pool"],
                "account": account,
                "userEmodeCategoryId": int(user_emode or 0),
                "positions": positions,
            }
        except Exception as exc:  # noqa: BLE001
            return False, str(exc)

    async def lend(
        self,
        *,
        underlying_token: str,
        qty: int,
        chain_id: int,
        native: bool = False,
    ) -> tuple[bool, Any]:
        strategy = self.strategy_wallet_address
        if not strategy:
            return False, "strategy wallet address not configured"
        qty = int(qty)
        if qty <= 0:
            return False, "qty must be positive"

        try:
            pool = await self._pool(chain_id=int(chain_id))

            if native:
                wrapped = await self._wrapped_native(chain_id=int(chain_id))
                wrap_tx = await encode_call(
                    target=wrapped,
                    abi=WETH_ABI,
                    fn_name="deposit",
                    args=[],
                    from_address=strategy,
                    chain_id=int(chain_id),
                    value=qty,
                )
                wrap_hash = await send_transaction(
                    wrap_tx, self.strategy_wallet_signing_callback
                )

                approved = await ensure_allowance(
                    token_address=wrapped,
                    owner=strategy,
                    spender=pool,
                    amount=qty,
                    chain_id=int(chain_id),
                    signing_callback=self.strategy_wallet_signing_callback,
                    approval_amount=MAX_UINT256,
                )
                if not approved[0]:
                    return approved

                supply_tx = await encode_call(
                    target=pool,
                    abi=POOL_ABI,
                    fn_name="supply",
                    args=[wrapped, qty, strategy, REFERRAL_CODE],
                    from_address=strategy,
                    chain_id=int(chain_id),
                )
                supply_hash = await send_transaction(
                    supply_tx, self.strategy_wallet_signing_callback
                )
                return True, {"wrap_tx": wrap_hash, "supply_tx": supply_hash}

            asset = to_checksum_address(underlying_token)
            approved = await ensure_allowance(
                token_address=asset,
                owner=strategy,
                spender=pool,
                amount=qty,
                chain_id=int(chain_id),
                signing_callback=self.strategy_wallet_signing_callback,
                approval_amount=MAX_UINT256,
            )
            if not approved[0]:
                return approved

            tx = await encode_call(
                target=pool,
                abi=POOL_ABI,
                fn_name="supply",
                args=[asset, qty, strategy, REFERRAL_CODE],
                from_address=strategy,
                chain_id=int(chain_id),
            )
            txn_hash = await send_transaction(tx, self.strategy_wallet_signing_callback)
            return True, txn_hash
        except Exception as exc:  # noqa: BLE001
            return False, str(exc)

    async def unlend(
        self,
        *,
        underlying_token: str,
        qty: int,
        chain_id: int,
        native: bool = False,
        withdraw_full: bool = False,
    ) -> tuple[bool, Any]:
        strategy = self.strategy_wallet_address
        if not strategy:
            return False, "strategy wallet address not configured"
        qty = int(qty)
        if qty <= 0 and not withdraw_full:
            return False, "qty must be positive"

        try:
            pool = await self._pool(chain_id=int(chain_id))
            amount = MAX_UINT256 if withdraw_full else qty

            if native:
                wrapped = await self._wrapped_native(chain_id=int(chain_id))
                before = await get_token_balance(
                    wrapped, int(chain_id), strategy, block_identifier="pending"
                )

                withdraw_tx = await encode_call(
                    target=pool,
                    abi=POOL_ABI,
                    fn_name="withdraw",
                    args=[wrapped, amount, strategy],
                    from_address=strategy,
                    chain_id=int(chain_id),
                )
                withdraw_hash = await send_transaction(
                    withdraw_tx, self.strategy_wallet_signing_callback
                )

                after = await get_token_balance(
                    wrapped, int(chain_id), strategy, block_identifier="pending"
                )
                unwrap_amount = max(0, int(after) - int(before))
                if unwrap_amount <= 0:
                    return True, {"withdraw_tx": withdraw_hash, "unwrap_tx": None}

                unwrap_tx = await encode_call(
                    target=wrapped,
                    abi=WETH_ABI,
                    fn_name="withdraw",
                    args=[int(unwrap_amount)],
                    from_address=strategy,
                    chain_id=int(chain_id),
                )
                unwrap_hash = await send_transaction(
                    unwrap_tx, self.strategy_wallet_signing_callback
                )
                return True, {"withdraw_tx": withdraw_hash, "unwrap_tx": unwrap_hash}

            asset = to_checksum_address(underlying_token)
            tx = await encode_call(
                target=pool,
                abi=POOL_ABI,
                fn_name="withdraw",
                args=[asset, amount, strategy],
                from_address=strategy,
                chain_id=int(chain_id),
            )
            txn_hash = await send_transaction(tx, self.strategy_wallet_signing_callback)
            return True, txn_hash
        except Exception as exc:  # noqa: BLE001
            return False, str(exc)

    async def borrow(
        self,
        *,
        underlying_token: str,
        qty: int,
        chain_id: int,
        native: bool = False,
    ) -> tuple[bool, Any]:
        strategy = self.strategy_wallet_address
        if not strategy:
            return False, "strategy wallet address not configured"
        qty = int(qty)
        if qty <= 0:
            return False, "qty must be positive"

        try:
            pool = await self._pool(chain_id=int(chain_id))
            asset = to_checksum_address(underlying_token)
            if native:
                wrapped = await self._wrapped_native(chain_id=int(chain_id))
                borrow_tx = await encode_call(
                    target=pool,
                    abi=POOL_ABI,
                    fn_name="borrow",
                    args=[wrapped, qty, VARIABLE_RATE_MODE, REFERRAL_CODE, strategy],
                    from_address=strategy,
                    chain_id=int(chain_id),
                )
                borrow_hash = await send_transaction(
                    borrow_tx, self.strategy_wallet_signing_callback
                )

                unwrap_tx = await encode_call(
                    target=wrapped,
                    abi=WETH_ABI,
                    fn_name="withdraw",
                    args=[qty],
                    from_address=strategy,
                    chain_id=int(chain_id),
                )
                unwrap_hash = await send_transaction(
                    unwrap_tx, self.strategy_wallet_signing_callback
                )
                return True, {"borrow_tx": borrow_hash, "unwrap_tx": unwrap_hash}

            tx = await encode_call(
                target=pool,
                abi=POOL_ABI,
                fn_name="borrow",
                args=[asset, qty, VARIABLE_RATE_MODE, REFERRAL_CODE, strategy],
                from_address=strategy,
                chain_id=int(chain_id),
            )
            txn_hash = await send_transaction(tx, self.strategy_wallet_signing_callback)
            return True, txn_hash
        except Exception as exc:  # noqa: BLE001
            return False, str(exc)

    async def repay(
        self,
        *,
        underlying_token: str,
        qty: int,
        chain_id: int,
        native: bool = False,
        repay_full: bool = False,
    ) -> tuple[bool, Any]:
        strategy = self.strategy_wallet_address
        if not strategy:
            return False, "strategy wallet address not configured"
        qty = int(qty)
        if qty <= 0 and not repay_full:
            return False, "qty must be positive"

        try:
            pool = await self._pool(chain_id=int(chain_id))

            if native:
                wrapped = await self._wrapped_native(chain_id=int(chain_id))
                repay_amount = MAX_UINT256 if repay_full else qty

                if repay_full:
                    # Read current variable debt (debt token balance) and wrap a small buffer.
                    v_debt = self._variable_debt_token_by_chain_underlying.get(
                        (int(chain_id), wrapped.lower())
                    )
                    if not v_debt:
                        ok, markets = await self.get_all_markets(
                            chain_id=int(chain_id), include_rewards=False
                        )
                        if not ok or not isinstance(markets, list):
                            return (
                                False,
                                f"failed to resolve reserves for chain_id={chain_id}",
                            )
                        for m in markets:
                            if (
                                str(m.get("underlying") or "").lower()
                                == wrapped.lower()
                            ):
                                v_debt = str(m.get("variable_debt_token") or "")
                                break
                        if not v_debt:
                            return (
                                False,
                                "could not resolve variable debt token for wrapped native",
                            )
                        self._variable_debt_token_by_chain_underlying[
                            (int(chain_id), wrapped.lower())
                        ] = to_checksum_address(v_debt)

                    debt = await get_token_balance(
                        str(v_debt),
                        int(chain_id),
                        strategy,
                        block_identifier="pending",
                    )
                    if debt <= 0:
                        return True, None

                    native_balance = await get_token_balance(
                        None,
                        int(chain_id),
                        strategy,
                        block_identifier="pending",
                    )
                    buffer_wei = max(1, int(debt) // 10_000)  # 0.01%
                    value = int(debt) + buffer_wei
                    if native_balance < value:
                        if native_balance < int(debt):
                            return (
                                False,
                                f"insufficient native balance for repay_full (debt_wei={debt}, balance_wei={native_balance})",
                            )
                        value = int(debt)
                else:
                    value = qty

                wrap_tx = await encode_call(
                    target=wrapped,
                    abi=WETH_ABI,
                    fn_name="deposit",
                    args=[],
                    from_address=strategy,
                    chain_id=int(chain_id),
                    value=int(value),
                )
                wrap_hash = await send_transaction(
                    wrap_tx, self.strategy_wallet_signing_callback
                )

                approved = await ensure_allowance(
                    token_address=wrapped,
                    owner=strategy,
                    spender=pool,
                    amount=MAX_UINT256 if repay_full else int(value),
                    chain_id=int(chain_id),
                    signing_callback=self.strategy_wallet_signing_callback,
                    approval_amount=MAX_UINT256,
                )
                if not approved[0]:
                    return approved

                repay_tx = await encode_call(
                    target=pool,
                    abi=POOL_ABI,
                    fn_name="repay",
                    args=[wrapped, int(repay_amount), VARIABLE_RATE_MODE, strategy],
                    from_address=strategy,
                    chain_id=int(chain_id),
                )
                repay_hash = await send_transaction(
                    repay_tx, self.strategy_wallet_signing_callback
                )
                return True, {"wrap_tx": wrap_hash, "repay_tx": repay_hash}

            repay_amount = MAX_UINT256 if repay_full else qty
            asset = to_checksum_address(underlying_token)
            allowance_target = MAX_UINT256 if repay_full else qty
            approved = await ensure_allowance(
                token_address=asset,
                owner=strategy,
                spender=pool,
                amount=allowance_target,
                chain_id=int(chain_id),
                signing_callback=self.strategy_wallet_signing_callback,
                approval_amount=MAX_UINT256,
            )
            if not approved[0]:
                return approved

            tx = await encode_call(
                target=pool,
                abi=POOL_ABI,
                fn_name="repay",
                args=[asset, int(repay_amount), VARIABLE_RATE_MODE, strategy],
                from_address=strategy,
                chain_id=int(chain_id),
            )
            txn_hash = await send_transaction(tx, self.strategy_wallet_signing_callback)
            return True, txn_hash
        except Exception as exc:  # noqa: BLE001
            return False, str(exc)

    async def set_collateral(
        self,
        *,
        underlying_token: str,
        chain_id: int,
        use_as_collateral: bool = True,
    ) -> tuple[bool, Any]:
        strategy = self.strategy_wallet_address
        if not strategy:
            return False, "strategy wallet address not configured"

        try:
            pool = await self._pool(chain_id=int(chain_id))
            asset = to_checksum_address(underlying_token)
            tx = await encode_call(
                target=pool,
                abi=POOL_ABI,
                fn_name="setUserUseReserveAsCollateral",
                args=[asset, bool(use_as_collateral)],
                from_address=strategy,
                chain_id=int(chain_id),
            )
            txn_hash = await send_transaction(tx, self.strategy_wallet_signing_callback)
            return True, txn_hash
        except Exception as exc:  # noqa: BLE001
            return False, str(exc)

    async def remove_collateral(
        self,
        *,
        underlying_token: str,
        chain_id: int,
    ) -> tuple[bool, Any]:
        return await self.set_collateral(
            underlying_token=str(underlying_token),
            chain_id=int(chain_id),
            use_as_collateral=False,
        )

    async def claim_all_rewards(
        self,
        *,
        chain_id: int,
        assets: list[str] | None = None,
        to_address: str | None = None,
    ) -> tuple[bool, Any]:
        strategy = self.strategy_wallet_address
        if not strategy:
            return False, "strategy wallet address not configured"

        try:
            entry = self._entry(int(chain_id))
            rewards_controller = to_checksum_address(entry["rewards_controller"])
            provider_addr = to_checksum_address(entry["pool_addresses_provider"])
            to_addr = to_checksum_address(to_address) if to_address else strategy

            if not assets:
                # Derive incentivized token addresses from the incentives data provider.
                ui_incentives_addr = to_checksum_address(
                    entry["ui_incentive_data_provider"]
                )
                async with web3_utils.web3_from_chain_id(int(chain_id)) as web3:
                    ui_incentives = web3.eth.contract(
                        address=ui_incentives_addr,
                        abi=UI_INCENTIVE_DATA_PROVIDER_V3_ABI,
                    )
                    inc_rows = await ui_incentives.functions.getReservesIncentivesData(
                        provider_addr
                    ).call(block_identifier="pending")

                assets_set: set[str] = set()
                for row in inc_rows or []:
                    for i in (1, 2, 3):
                        try:
                            token_addr = str((row[i] or [None])[0] or "")
                        except Exception:
                            token_addr = ""
                        if (
                            token_addr
                            and token_addr
                            != "0x0000000000000000000000000000000000000000"
                        ):
                            assets_set.add(to_checksum_address(token_addr))
                assets = sorted(assets_set)

            if not assets:
                return True, {"claimed": [], "note": "no incentivized assets found"}

            tx = await encode_call(
                target=rewards_controller,
                abi=REWARDS_CONTROLLER_ABI,
                fn_name="claimAllRewards",
                args=[[to_checksum_address(a) for a in assets], to_addr],
                from_address=strategy,
                chain_id=int(chain_id),
            )
            txn_hash = await send_transaction(tx, self.strategy_wallet_signing_callback)
            return True, txn_hash
        except Exception as exc:  # noqa: BLE001
            return False, str(exc)



from __future__ import annotations

import os
from pathlib import Path

from wayfinder_paths.runner.paths import RunnerPaths


def runs_root(paths: RunnerPaths) -> Path:
    candidate = (os.getenv("WAYFINDER_RUNS_DIR") or ".wayfinder_runs").strip()
    p = Path(candidate)
    if not p.is_absolute():
        p = paths.repo_root / p
    return p.resolve(strict=False)


def resolve_script_path(paths: RunnerPaths, script_path: str) -> Path:
    raw = str(script_path).strip()
    if not raw:
        raise ValueError("script_path is required")

    p = Path(raw)
    if not p.is_absolute():
        p = paths.repo_root / p
    resolved = p.resolve(strict=False)

    root = runs_root(paths)
    try:
        resolved.relative_to(root)
    except ValueError as exc:
        raise ValueError(
            f"script_path must be inside the local runs directory ({root})"
        ) from exc

    if not resolved.exists():
        raise FileNotFoundError(f"Script file not found: {resolved}")

    if resolved.suffix.lower() != ".py":
        raise ValueError("Only .py scripts are supported")

    return resolved



from wayfinder_paths.core.constants.contracts import ENSO_ROUTER
from wayfinder_paths.policies.util import allow_functions


async def enso_swap():
    return await allow_functions(
        policy_name="Allow Enso Swap",
        abi_chain_id=8453,
        address=ENSO_ROUTER,
        function_names=[
            "routeMulti",
            "routeSingle",
            "safeRouteMulti",
            "safeRouteSingle",
        ],
    )



from __future__ import annotations

# Minimal ABIs for rewards claiming.

MERKL_DISTRIBUTOR_ABI = [
    {
        "type": "function",
        "stateMutability": "nonpayable",
        "name": "claim",
        "inputs": [
            {"name": "users", "type": "address[]"},
            {"name": "tokens", "type": "address[]"},
            {"name": "amounts", "type": "uint256[]"},
            {"name": "proofs", "type": "bytes32[][]"},
        ],
        "outputs": [],
    }
]

URD_ABI = [
    {
        "type": "function",
        "stateMutability": "nonpayable",
        "name": "claim",
        "inputs": [
            {"name": "account", "type": "address"},
            {"name": "reward", "type": "address"},
            {"name": "claimable", "type": "uint256"},
            {"name": "proof", "type": "bytes32[]"},
        ],
        "outputs": [{"name": "amount", "type": "uint256"}],
    }
]



from __future__ import annotations

from pathlib import Path

import pytest

from wayfinder_paths.mcp.tools.run_script import run_script


@pytest.mark.asyncio
async def test_run_script_rejects_outside_runs_dir(tmp_path: Path, monkeypatch):
    runs_root = tmp_path / "runs"
    runs_root.mkdir()
    monkeypatch.setenv("WAYFINDER_RUNS_DIR", str(runs_root))
    monkeypatch.setenv("WAYFINDER_MCP_STATE_PATH", str(tmp_path / "mcp.sqlite3"))

    outside = tmp_path / "outside.py"
    outside.write_text("print('nope')\n")

    out = await run_script(script_path=str(outside))
    assert out["ok"] is False
    assert out["error"]["code"] == "invalid_request"


@pytest.mark.asyncio
async def test_run_script_executes(tmp_path: Path, monkeypatch):
    runs_root = tmp_path / "runs"
    runs_root.mkdir()
    monkeypatch.setenv("WAYFINDER_RUNS_DIR", str(runs_root))
    monkeypatch.setenv("WAYFINDER_MCP_STATE_PATH", str(tmp_path / "mcp.sqlite3"))

    script = runs_root / "hello.py"
    script.write_text("import os\nprint('PWD=' + os.getcwd())\n")

    out1 = await run_script(script_path=str(script), timeout_s=30)
    assert out1["ok"] is True
    assert out1["result"]["exit_code"] == 0



from __future__ import annotations

from dataclasses import dataclass, field
from datetime import datetime
from enum import Enum, auto
from typing import Any

from .constants import (
    ALLOCATION_DEVIATION_THRESHOLD,
    BOROS_ENABLE_MIN_TOTAL_USD,
    FULL_REBALANCE_THRESHOLD,
    PARTIAL_TRIM_THRESHOLD,
)

# ─────────────────────────────────────────────────────────────────────────────
# CONFIGURATION
# ─────────────────────────────────────────────────────────────────────────────


@dataclass
class HedgeConfig:
    spot_pct: float  # Total allocation to spot (kHYPE + looped HYPE)
    khype_fraction: float  # Fraction of spot allocation for kHYPE
    looped_hype_fraction: float  # Fraction of spot allocation for looped HYPE
    hyperliquid_pct: float  # Allocation to Hyperliquid perp margin
    boros_pct: float  # Allocation to Boros rate lock

    @classmethod
    def default(cls) -> HedgeConfig:
        return cls(
            # New routing:
            # - Send 100% of Arbitrum USDC to Hyperliquid first.
            # - Target ~50% of portfolio value staying on Hyperliquid (margin + cash buffer).
            # - Target ~50% HYPE exposure split between HyperEVM spot (≈45%) and Boros HYPE collateral (≈5%).
            spot_pct=0.45,
            khype_fraction=0.5,
            looped_hype_fraction=0.5,
            hyperliquid_pct=0.50,
            boros_pct=0.05,
        )


# ─────────────────────────────────────────────────────────────────────────────
# INVENTORY (World Model)
# ─────────────────────────────────────────────────────────────────────────────


@dataclass
class Inventory:
    # HyperEVM wallet balances
    hype_hyperevm_balance: float
    hype_hyperevm_value_usd: float
    whype_balance: float  # Wrapped HYPE balance on HyperEVM
    whype_value_usd: float  # Value in USD (1:1 with HYPE)
    khype_balance: float
    khype_value_usd: float
    looped_hype_balance: float
    looped_hype_value_usd: float

    # Arbitrum wallet balances
    usdc_arb_idle: float
    usdt_arb_idle: float
    eth_arb_balance: float  # For gas

    # Arbitrum OFT HYPE (bridged from HyperEVM)
    hype_oft_arb_balance: float
    hype_oft_arb_value_usd: float

    # Hyperliquid venue
    hl_perp_margin: float
    hl_spot_usdc: float
    hl_spot_hype: float
    hl_spot_hype_value_usd: float
    hl_short_size_hype: float
    hl_short_value_usd: float
    hl_unrealized_pnl: float
    hl_withdrawable_usd: float

    # Boros venue
    boros_idle_collateral_isolated: float  # HYPE units
    boros_idle_collateral_cross: float  # HYPE units
    boros_collateral_hype: float  # Total HYPE collateral in Boros
    boros_collateral_usd: float  # USD value (HYPE * hype_price_usd)
    boros_pending_withdrawal_hype: float  # HYPE units
    boros_pending_withdrawal_usd: float  # USD value
    boros_committed_collateral_usd: float  # boros_collateral_usd + hype_oft_arb_value_usd (+ any in-flight OFT bridge)
    boros_position_size: float  # YU notional
    boros_position_value: float  # Unrealized PnL

    # Exchange rates
    khype_to_hype_ratio: float
    looped_hype_to_hype_ratio: float
    hype_price_usd: float

    # Aggregates
    spot_value_usd: float
    total_hype_exposure: float
    total_value: float

    # Optional
    boros_position_market_ids: list[int] | None = None

    # Liquidation detection
    hl_liquidation_detected: bool = False
    hl_liquidation_fills: list[dict] = field(default_factory=list)


@dataclass
class AllocationStatus:
    # Actual values
    spot_value: float
    hl_value: float
    boros_value: float
    idle_value: float
    total_value: float

    # Actual percentages
    spot_pct_actual: float
    hl_pct_actual: float
    boros_pct_actual: float

    # Deviations from target (negative = underallocated)
    spot_deviation: float
    hl_deviation: float
    boros_deviation: float

    # Dollar amounts needed to reach target
    spot_needed_usd: float
    hl_needed_usd: float
    boros_needed_usd: float


@dataclass
class YieldInfo:
    khype_apy: float | None = None
    lhype_apy: float | None = None
    boros_apr: float | None = None
    hl_funding_rate: float | None = None

    khype_expected_yield_usd: float = 0.0
    lhype_expected_yield_usd: float = 0.0
    boros_expected_yield_usd: float = 0.0
    hl_expected_yield_usd: float = 0.0

    total_expected_yield_usd: float = 0.0
    blended_apy: float | None = None


# ─────────────────────────────────────────────────────────────────────────────
# PLANNING ENUMS
# ─────────────────────────────────────────────────────────────────────────────


class Mode(Enum):
    NORMAL = auto()  # Regular operations
    TRIM = auto()  # Risk at 75%+: reduce exposure
    REDEPLOY = auto()  # Risk at 90%+: emergency redeploy


class PlanOp(Enum):
    # Priority 0: Safety/Risk mitigation
    CLOSE_AND_REDEPLOY = "close_and_redeploy"
    PARTIAL_TRIM_SPOT = "partial_trim_spot"
    COMPLETE_PENDING_WITHDRAWAL = "complete_pending_withdrawal"

    # Priority 5: Gas routing (must happen first!)
    ENSURE_GAS_ON_HYPEREVM = "ensure_gas_on_hyperevm"
    ENSURE_GAS_ON_ARBITRUM = "ensure_gas_on_arbitrum"

    # Priority 10: Capital routing
    FUND_BOROS = "fund_boros"
    SEND_USDC_TO_HL = "send_usdc_to_hl"
    BRIDGE_TO_HYPEREVM = "bridge_to_hyperevm"
    TRANSFER_HL_SPOT_TO_HYPEREVM = "transfer_hl_spot_to_hyperevm"
    DEPLOY_EXCESS_HL_MARGIN = "deploy_excess_hl_margin"

    # Priority 20: Position management
    SWAP_HYPE_TO_LST = "swap_hype_to_lst"
    ENSURE_HL_SHORT = "ensure_hl_short"

    # Priority 30: Rate positions
    ENSURE_BOROS_POSITION = "ensure_boros_position"


# ─────────────────────────────────────────────────────────────────────────────
# PLANNING DATACLASSES
# ─────────────────────────────────────────────────────────────────────────────


@dataclass
class PlanStep:
    op: PlanOp
    priority: int  # Lower = execute first
    key: str  # Unique identifier for deduplication
    params: dict[str, Any] = field(default_factory=dict)
    reason: str = ""

    def __lt__(self, other: PlanStep) -> bool:
        return self.priority < other.priority


@dataclass
class PlannerConfig:
    # Risk thresholds
    partial_trim_threshold: float = PARTIAL_TRIM_THRESHOLD
    full_rebalance_threshold: float = FULL_REBALANCE_THRESHOLD

    # Allocation thresholds
    allocation_deviation_threshold: float = ALLOCATION_DEVIATION_THRESHOLD
    position_size_tolerance: float = 0.05

    # Minimum amounts to act on
    min_usdc_action: float = 5.0
    min_usdt_action: float = 1.0
    min_hype_swap: float = 0.1

    # Boros guard
    min_total_for_boros: float = BOROS_ENABLE_MIN_TOTAL_USD

    # Hyperliquid guardrails
    hl_withdrawable_buffer_usd: float = 5.0
    hl_withdraw_for_boros_cooldown_minutes: int = 30
    hl_max_withdraw_for_boros_usd: float = 25.0

    # HL margin management
    hl_target_margin_ratio: float = 0.50
    hl_margin_buffer_ratio: float = 0.15

    # Boros hysteresis
    boros_market_switch_cooldown_hours: int = 24
    boros_apr_improvement_threshold: float = 0.02

    # Boros coverage - target 100%, resize threshold is the hysteresis band
    boros_coverage_target: float = 1.0
    boros_resize_min_excess_usd: float = (
        10.0  # Min YU diff to trigger resize (hysteresis)
    )

    # Delta neutral tolerances
    delta_neutral_rel_tol: float = 0.02
    delta_neutral_abs_tol_hype: float = 0.11

    # Execution limits
    max_steps_per_iteration: int = 5
    max_iterations_per_tick: int = 4
    max_total_steps_per_tick: int = 15

    @classmethod
    def default(cls) -> PlannerConfig:
        return cls()


@dataclass
class PlannerRuntime:
    # Boros market selection (persistent state for hysteresis)
    current_boros_market_id: int | None = None
    current_boros_token_id: int | None = None
    current_boros_collateral_address: str | None = None
    boros_market_selected_at: datetime | None = None

    # Execution tracking
    last_update_at: datetime | None = None
    steps_executed_this_session: int = 0

    # Virtual ledger for same-tick tracking
    committed_usdc_arb: float = 0.0
    committed_usdt_arb: float = 0.0
    funded_boros_this_tick: bool = False

    last_hl_withdraw_for_boros_at: datetime | None = None

    # OFT bridge tracking (HyperEVM native HYPE -> Arbitrum OFT HYPE) so we don't
    # repeatedly fund Boros while the bridge is still settling.
    in_flight_boros_oft_hype: float = 0.0
    in_flight_boros_oft_hype_balance_before: float = 0.0
    in_flight_boros_oft_hype_started_at: datetime | None = None

    # HL state
    leverage_set_for_hype: bool = False

    def reset_virtual_ledger(self) -> None:
        self.committed_usdc_arb = 0.0
        self.committed_usdt_arb = 0.0

    def reset_tick_flags(self) -> None:
        self.funded_boros_this_tick = False

    def available_usdc_arb(self, inv_usdc: float) -> float:
        return max(0.0, inv_usdc - self.committed_usdc_arb)

    def available_usdt_arb(self, inv_usdt: float) -> float:
        return max(0.0, inv_usdt - self.committed_usdt_arb)

    def commit_usdc(self, amount: float) -> None:
        self.committed_usdc_arb += amount

    def commit_usdt(self, amount: float) -> None:
        self.committed_usdt_arb += amount


@dataclass
class DesiredState:
    mode: Mode

    # Target allocations (USD)
    target_spot_usd: float
    target_hl_margin_usd: float
    target_boros_collateral_usd: float

    # Target positions
    target_hype_short_size: float
    # Boros order sizing is in YU. YU units depend on collateral:
    # - HYPE collateral: 1 YU = 1 HYPE
    # - USDT collateral: 1 YU = 1 USDT (≈$1)
    target_boros_position_yu: float

    # Selected Boros market
    boros_market_id: int | None = None
    boros_market_symbol: str | None = None
    boros_tenor_days: float | None = None


@dataclass
class Plan:
    desired_state: DesiredState
    steps: list[PlanStep] = field(default_factory=list)
    messages: list[str] = field(default_factory=list)

    def add_step(
        self,
        op: PlanOp,
        priority: int,
        key: str,
        params: dict[str, Any] | None = None,
        reason: str = "",
    ) -> None:
        if not any(s.key == key for s in self.steps):
            self.steps.append(
                PlanStep(
                    op=op,
                    priority=priority,
                    key=key,
                    params=params or {},
                    reason=reason,
                )
            )

    def sort_steps(self) -> None:
        self.steps.sort()


# Operations that change inventory and require re-observing
# Note: ENSURE_GAS_ON_HYPEREVM is excluded because it may return early
# without actually changing anything (e.g., "will be provisioned during routing")
INVENTORY_CHANGING_OPS: set[PlanOp] = {
    PlanOp.CLOSE_AND_REDEPLOY,
    PlanOp.PARTIAL_TRIM_SPOT,
    PlanOp.COMPLETE_PENDING_WITHDRAWAL,
    PlanOp.FUND_BOROS,
    PlanOp.SEND_USDC_TO_HL,
    PlanOp.BRIDGE_TO_HYPEREVM,
    PlanOp.TRANSFER_HL_SPOT_TO_HYPEREVM,
    PlanOp.DEPLOY_EXCESS_HL_MARGIN,
    PlanOp.SWAP_HYPE_TO_LST,
    PlanOp.ENSURE_HL_SHORT,
    PlanOp.ENSURE_BOROS_POSITION,
}



"""
Risk and recovery operations for BorosHypeStrategy.

Kept as a mixin so the main strategy file stays readable without changing behavior.
"""

from __future__ import annotations

import asyncio
from typing import Any

from loguru import logger

from wayfinder_paths.adapters.hyperliquid_adapter.adapter import HyperliquidAdapter
from wayfinder_paths.adapters.hyperliquid_adapter.paired_filler import PairedFiller

from .constants import (
    BOROS_HYPE_MARKET_ID,
    BOROS_HYPE_TOKEN_ID,
    HYPE_NATIVE,
    KHYPE_LST,
    LOOPED_HYPE,
    MIN_HYPE_GAS,
    MIN_NET_DEPOSIT,
    USDC_ARB,
    USDT_ARB,
    WHYPE,
)
from .types import Inventory


class BorosHypeRiskOpsMixin:
    async def _close_and_redeploy(
        self, params: dict[str, Any], inventory: Inventory
    ) -> tuple[bool, str]:
        ok, msg = self._require_adapters(
            "balance_adapter", "hyperliquid_adapter", "brap_adapter"
        )
        if not ok:
            return False, msg
        if not self._sign_callback:
            return False, "No strategy wallet signing callback configured"

        ok_addr, address = self._require_strategy_wallet_address()
        if not ok_addr:
            return False, address
        strategy_wallet = self._config.get("strategy_wallet", {})

        logger.warning("Emergency close and redeploy triggered")

        try:
            await self._cancel_hl_open_orders_for_hype(address)
        except Exception as exc:  # noqa: BLE001
            logger.debug(f"Failed to cancel HL open orders pre-redeploy: {exc}")

        try:
            ok_state, user_state = await self.hyperliquid_adapter.get_user_state(
                address
            )
            current_short_size = 0.0
            if ok_state and isinstance(user_state, dict):
                for pos in user_state.get("assetPositions", []):
                    p = pos.get("position", {}) if isinstance(pos, dict) else {}
                    if p.get("coin") == "HYPE":
                        szi = float(p.get("szi", 0))
                        if szi < 0:
                            current_short_size = abs(szi)
                        break

            if current_short_size > 0.01:
                perp_asset_id = self.hyperliquid_adapter.coin_to_asset.get("HYPE")
                if perp_asset_id is None:
                    logger.warning(
                        "Missing Hyperliquid perp asset id for HYPE; cannot close hedge"
                    )
                else:
                    rounded_size = self.hyperliquid_adapter.get_valid_order_size(
                        int(perp_asset_id), current_short_size
                    )
                    if rounded_size > 0:
                        (
                            ok_close,
                            res_close,
                        ) = await self.hyperliquid_adapter.place_market_order(
                            asset_id=int(perp_asset_id),
                            is_buy=True,
                            slippage=0.05,
                            size=float(rounded_size),
                            address=address,
                            reduce_only=True,
                            builder=self.builder_fee,
                        )
                        if not ok_close:
                            logger.warning(f"Failed to close HL short: {res_close}")
                        await asyncio.sleep(2)
        except Exception as exc:  # noqa: BLE001
            logger.warning(f"Failed closing HL hedge: {exc}")

        if self.boros_adapter:
            try:
                ok_pos, positions = await self.boros_adapter.get_active_positions()
                if ok_pos and isinstance(positions, list):
                    for pos in positions:
                        mid = pos.get("marketId") or pos.get("market_id")
                        try:
                            mid_int = int(mid) if mid is not None else None
                        except (TypeError, ValueError):
                            mid_int = None
                        if mid_int and mid_int > 0:
                            try:
                                await self.boros_adapter.close_positions_market(mid_int)
                            except Exception as exc:  # noqa: BLE001
                                logger.warning(
                                    f"Failed to close Boros market {mid_int}: {exc}"
                                )
            except Exception as exc:  # noqa: BLE001
                logger.warning(f"Failed to close Boros positions: {exc}")

        try:
            ok_khype, khype_raw = await self.balance_adapter.get_vault_wallet_balance(
                KHYPE_LST
            )
            if ok_khype and khype_raw > 0:
                await self.brap_adapter.swap_from_token_ids(
                    from_token_id=KHYPE_LST,
                    to_token_id=HYPE_NATIVE,
                    from_address=address,
                    amount=str(int(khype_raw)),
                    slippage=0.01,
                    strategy_name="boros_hype_strategy",
                )

            ok_lhype, lhype_raw = await self.balance_adapter.get_vault_wallet_balance(
                LOOPED_HYPE
            )
            if ok_lhype and lhype_raw > 0:
                await self.brap_adapter.swap_from_token_ids(
                    from_token_id=LOOPED_HYPE,
                    to_token_id=HYPE_NATIVE,
                    from_address=address,
                    amount=str(int(lhype_raw)),
                    slippage=0.01,
                    strategy_name="boros_hype_strategy",
                )
        except Exception as exc:  # noqa: BLE001
            logger.warning(f"Failed selling spot LSTs to HYPE: {exc}")

        try:
            ok_hype, hype_raw = await self.balance_adapter.get_vault_wallet_balance(
                HYPE_NATIVE
            )
            hype_raw_int = int(hype_raw) if ok_hype and hype_raw and hype_raw > 0 else 0
            gas_reserve_wei = int(float(MIN_HYPE_GAS) * 1e18)
            hype_to_transfer_raw = max(0, hype_raw_int - gas_reserve_wei)

            if hype_to_transfer_raw > int(0.01 * 1e18):
                destination = HyperliquidAdapter.hypercore_index_to_system_address(150)
                ok_send, send_res = await self.balance_adapter.send_to_address(
                    token_id=HYPE_NATIVE,
                    amount=int(hype_to_transfer_raw),
                    from_wallet=strategy_wallet,
                    to_address=destination,
                    signing_callback=self._sign_callback,
                )
                if not ok_send:
                    logger.warning(
                        f"Failed to transfer HYPE to Hyperliquid: {send_res}"
                    )
                await asyncio.sleep(3)
        except Exception as exc:  # noqa: BLE001
            logger.warning(f"Failed to transfer HYPE to Hyperliquid: {exc}")

        try:
            ok_spot, spot_state = await self.hyperliquid_adapter.get_spot_user_state(
                address
            )
            spot_hype_balance = 0.0
            if ok_spot and isinstance(spot_state, dict):
                for bal in spot_state.get("balances", []):
                    token = bal.get("coin") or bal.get("token")
                    hold = float(bal.get("hold", 0))
                    total = float(bal.get("total", 0))
                    available = total - hold
                    if token == "HYPE":
                        spot_hype_balance = available
                        break

            if spot_hype_balance > 0.01:
                spot_asset_id, _ = await self._get_hype_asset_ids()
                rounded_size = self.hyperliquid_adapter.get_valid_order_size(
                    int(spot_asset_id), spot_hype_balance
                )
                if rounded_size > 0:
                    await self.hyperliquid_adapter.place_market_order(
                        asset_id=int(spot_asset_id),
                        is_buy=False,
                        slippage=0.10,
                        size=float(rounded_size),
                        address=address,
                        builder=self.builder_fee,
                    )
                    await asyncio.sleep(2)
        except Exception as exc:  # noqa: BLE001
            logger.warning(f"Failed to sell spot HYPE: {exc}")

        try:
            usdc_sz_decimals = 8

            ok_spot, spot_state = await self.hyperliquid_adapter.get_spot_user_state(
                address
            )
            spot_usdc = 0.0
            spot_total_s = "0"
            spot_hold_s = "0"
            if ok_spot and isinstance(spot_state, dict):
                for bal in spot_state.get("balances", []):
                    token = bal.get("coin") or bal.get("token")
                    if token != "USDC":
                        continue
                    spot_total_s = str(bal.get("total", "0") or "0")
                    spot_hold_s = str(bal.get("hold", "0") or "0")
                    hold = float(spot_hold_s)
                    total = float(spot_total_s)
                    spot_usdc = max(0.0, total - hold)
                    break
            if spot_usdc > 1.0:
                amount = self.hyperliquid_adapter.max_transferable_amount(
                    spot_total_s,
                    spot_hold_s,
                    sz_decimals=int(usdc_sz_decimals),
                    leave_one_tick=True,
                )
                (
                    ok_xfer,
                    res_xfer,
                ) = await self.hyperliquid_adapter.transfer_spot_to_perp(
                    amount=float(amount),
                    address=address,
                )
                if (not ok_xfer) and int(usdc_sz_decimals) != 2:
                    if "insufficient balance" in str(res_xfer).lower():
                        fallback_2dp = self.hyperliquid_adapter.max_transferable_amount(
                            spot_total_s,
                            spot_hold_s,
                            sz_decimals=2,
                            leave_one_tick=True,
                        )
                        if fallback_2dp > 1.0:
                            await self.hyperliquid_adapter.transfer_spot_to_perp(
                                amount=float(fallback_2dp),
                                address=address,
                            )
        except Exception as exc:  # noqa: BLE001
            logger.warning(f"Failed to move spot USDC to perp: {exc}")

        hl_perp_balance = 0.0
        try:
            ok_state, user_state = await self.hyperliquid_adapter.get_user_state(
                address
            )
            if ok_state and isinstance(user_state, dict):
                hl_perp_balance = float(
                    self.hyperliquid_adapter.get_perp_margin_amount(user_state)
                )
        except Exception as exc:  # noqa: BLE001
            logger.warning(f"Failed to read HL perp balance: {exc}")

        ok_arb_usdc, arb_usdc_raw = await self.balance_adapter.get_vault_wallet_balance(
            USDC_ARB, wallet_address=address
        )
        arb_usdc_tokens = (
            (int(arb_usdc_raw) / 1e6) if ok_arb_usdc and arb_usdc_raw else 0.0
        )
        total_usdc = hl_perp_balance + arb_usdc_tokens

        if total_usdc < MIN_NET_DEPOSIT:
            return True, "Closed all positions. Insufficient capital to redeploy."

        spot_target = self.hedge_cfg.spot_pct * total_usdc
        boros_target = self.hedge_cfg.boros_pct * total_usdc

        if spot_target > self._planner_config.min_usdc_action:
            try:
                await self.hyperliquid_adapter.transfer_perp_to_spot(
                    amount=float(spot_target),
                    address=address,
                )
                await asyncio.sleep(2)

                success, mids = await self.hyperliquid_adapter.get_all_mid_prices()
                hype_price = (
                    float(mids.get("HYPE", 0.0))
                    if success and isinstance(mids, dict)
                    else 0.0
                )
                if hype_price <= 0:
                    hype_price = float(inventory.hype_price_usd or 0.0)

                if hype_price > 0:
                    hype_to_buy = float(spot_target) / hype_price
                    spot_asset_id, _ = await self._get_hype_asset_ids()
                    rounded_size = self.hyperliquid_adapter.get_valid_order_size(
                        int(spot_asset_id), hype_to_buy
                    )
                    if rounded_size > 0:
                        await self.hyperliquid_adapter.place_market_order(
                            asset_id=int(spot_asset_id),
                            is_buy=True,
                            slippage=0.10,
                            size=float(rounded_size),
                            address=address,
                            builder=self.builder_fee,
                        )
                        await asyncio.sleep(2)

                        (
                            ok_spot,
                            spot_state,
                        ) = await self.hyperliquid_adapter.get_spot_user_state(address)
                        spot_hype = 0.0
                        if ok_spot and isinstance(spot_state, dict):
                            for bal in spot_state.get("balances", []):
                                token = bal.get("coin") or bal.get("token")
                                if token == "HYPE":
                                    hold = float(bal.get("hold", 0))
                                    total = float(bal.get("total", 0))
                                    spot_hype = max(0.0, total - hold)
                                    break

                        amount_to_bridge = spot_hype - 0.001
                        if amount_to_bridge > 0.1:
                            await self.hyperliquid_adapter.hypercore_to_hyperevm(
                                amount=float(amount_to_bridge),
                                address=address,
                            )
            except Exception as exc:  # noqa: BLE001
                logger.warning(f"Failed to redeploy spot: {exc}")

        if boros_target > self._planner_config.min_usdt_action:
            try:
                ok_wd, wd_res = await self.hyperliquid_adapter.withdraw(
                    amount=float(boros_target),
                    address=address,
                )
                if ok_wd:
                    await self.hyperliquid_adapter.wait_for_withdrawal(
                        address=address,
                        max_poll_time_s=300,
                        poll_interval_s=10,
                    )
                    (
                        ok_arb,
                        arb_raw,
                    ) = await self.balance_adapter.get_vault_wallet_balance(
                        USDC_ARB, wallet_address=address
                    )
                    if ok_arb and arb_raw and int(arb_raw) > 0:
                        await self.brap_adapter.swap_from_token_ids(
                            from_token_id=USDC_ARB,
                            to_token_id=USDT_ARB,
                            from_address=address,
                            amount=str(int(arb_raw)),
                            slippage=0.005,
                            strategy_name="boros_hype_strategy",
                        )
                else:
                    logger.warning(
                        f"Failed to withdraw from HL for Boros redeploy: {wd_res}"
                    )
            except Exception as exc:  # noqa: BLE001
                logger.warning(f"Failed to redeploy Boros: {exc}")

        try:
            inv_after = await self.observe()
            swappable_hype = max(
                0.0, float(inv_after.hype_hyperevm_balance or 0.0) - MIN_HYPE_GAS
            )
            if swappable_hype > self._planner_config.min_hype_swap:
                await self._swap_hype_to_lst({"hype_amount": swappable_hype}, inv_after)
        except Exception as exc:  # noqa: BLE001
            logger.warning(f"Failed to allocate spot HYPE after redeploy: {exc}")

        try:
            inv_final = await self.observe()
            ok_short, msg_short = await self._ensure_hl_short(
                {
                    "target_size": inv_final.total_hype_exposure,
                    "current_size": inv_final.hl_short_size_hype,
                },
                inv_final,
            )
            if not ok_short:
                logger.error(
                    f"[RECOVERY_FAIL] Could not re-open HL hedge after redeploy: {msg_short}"
                )
                return await self._failsafe_liquidate_all(
                    f"Post-liquidation hedge rebuild failed: {msg_short}"
                )

            # If we cannot get Boros back into a sane state, melt down rather than
            # continuing with a partially functioning multi-venue position.
            try:
                spot_usd = float(inv_final.total_hype_exposure) * float(
                    inv_final.hype_price_usd
                )
                boros_enabled = (
                    float(inv_final.total_value)
                    >= float(self._planner_config.min_total_for_boros)
                    and float(inv_final.boros_pending_withdrawal_usd) <= 0.0
                )
                if boros_enabled and spot_usd >= 10.0 and self.boros_adapter:
                    market_id = (
                        self._planner_runtime.current_boros_market_id
                        or BOROS_HYPE_MARKET_ID
                    )
                    token_id = int(
                        self._planner_runtime.current_boros_token_id
                        or BOROS_HYPE_TOKEN_ID
                    )
                    if token_id == BOROS_HYPE_TOKEN_ID:
                        target_yu = float(inv_final.total_hype_exposure) * float(
                            self._planner_config.boros_coverage_target
                        )
                    else:
                        target_yu = spot_usd * float(
                            self._planner_config.boros_coverage_target
                        )
                    ok_boros, msg_boros = await self._ensure_boros_position(
                        {
                            "market_id": int(market_id),
                            "target_size_yu": float(target_yu),
                        },
                        inv_final,
                    )
                    if not ok_boros:
                        return await self._failsafe_liquidate_all(
                            f"Post-liquidation Boros recovery failed: {msg_boros}"
                        )
            except Exception as exc:  # noqa: BLE001
                return await self._failsafe_liquidate_all(
                    f"Post-liquidation Boros recovery raised: {exc}"
                )

            return (
                True,
                f"Redeployed. Spot={inv_final.total_hype_exposure:.4f} HYPE, short re-opened.",
            )
        except Exception as exc:  # noqa: BLE001
            logger.warning(f"Failed to verify hedge after redeploy: {exc}")
            return True, "Redeployed (hedge verification pending)"

    async def _failsafe_liquidate_all(self, reason: str) -> tuple[bool, str]:
        # Called when critical operations fail; close all positions to stable assets
        logger.error(f"[FAILSAFE] Initiating full liquidation: {reason}")
        self._failsafe_triggered = True

        messages: list[str] = []

        ok_addr, address = self._require_strategy_wallet_address()
        if not ok_addr:
            msg = f"[FAILSAFE] No wallet address: {reason}"
            self._failsafe_message = msg
            return False, msg
        if not self._sign_callback:
            msg = f"[FAILSAFE] No signing callback: {reason}"
            self._failsafe_message = msg
            return False, msg
        strategy_wallet = self._config.get("strategy_wallet", {})

        if self.hyperliquid_adapter:
            try:
                ok_state, user_state = await self.hyperliquid_adapter.get_user_state(
                    address
                )
                current_short_size = 0.0
                if ok_state and isinstance(user_state, dict):
                    for pos in user_state.get("assetPositions", []):
                        p = pos.get("position", {}) if isinstance(pos, dict) else {}
                        if p.get("coin") == "HYPE":
                            szi = float(p.get("szi", 0))
                            if szi < 0:
                                current_short_size = abs(szi)
                            break

                if current_short_size > 0.01:
                    perp_asset_id = self.hyperliquid_adapter.coin_to_asset.get("HYPE")
                    if perp_asset_id is not None:
                        rounded_size = self.hyperliquid_adapter.get_valid_order_size(
                            int(perp_asset_id), current_short_size
                        )
                        if rounded_size > 0:
                            (
                                ok_close,
                                res_close,
                            ) = await self.hyperliquid_adapter.place_market_order(
                                asset_id=int(perp_asset_id),
                                is_buy=True,
                                slippage=0.05,
                                size=float(rounded_size),
                                address=address,
                                reduce_only=True,
                                builder=self.builder_fee,
                            )
                            if ok_close:
                                messages.append(f"HL short closed: {rounded_size:.4f}")
                            else:
                                messages.append(f"HL close failed: {res_close}")
                            await asyncio.sleep(2)
                else:
                    messages.append("HL short: none")
            except Exception as e:
                messages.append(f"HL close error: {e}")

        if self.boros_adapter:
            try:
                ok_pos, positions = await self.boros_adapter.get_active_positions()
                if ok_pos and isinstance(positions, list) and positions:
                    for pos in positions:
                        mid = pos.get("marketId") or pos.get("market_id")
                        try:
                            mid_int = int(mid) if mid is not None else None
                        except (TypeError, ValueError):
                            mid_int = None
                        if mid_int and mid_int > 0:
                            try:
                                await self.boros_adapter.close_positions_market(mid_int)
                                messages.append(f"Boros {mid_int} closed")
                            except Exception as exc:
                                messages.append(f"Boros {mid_int} close failed: {exc}")
                else:
                    messages.append("Boros positions: none")
            except Exception as e:
                messages.append(f"Boros close error: {e}")

        if self.brap_adapter and self.balance_adapter:
            try:
                # Swap kHYPE to HYPE
                (
                    ok_khype,
                    khype_raw,
                ) = await self.balance_adapter.get_vault_wallet_balance(KHYPE_LST)
                if ok_khype and khype_raw and int(khype_raw) > 0:
                    await self.brap_adapter.swap_from_token_ids(
                        from_token_id=KHYPE_LST,
                        to_token_id=HYPE_NATIVE,
                        from_address=address,
                        amount=str(int(khype_raw)),
                        slippage=0.02,
                        strategy_name="boros_hype_strategy",
                    )
                    messages.append("kHYPE swapped to HYPE")

                # Swap lHYPE to HYPE
                (
                    ok_lhype,
                    lhype_raw,
                ) = await self.balance_adapter.get_vault_wallet_balance(LOOPED_HYPE)
                if ok_lhype and lhype_raw and int(lhype_raw) > 0:
                    await self.brap_adapter.swap_from_token_ids(
                        from_token_id=LOOPED_HYPE,
                        to_token_id=HYPE_NATIVE,
                        from_address=address,
                        amount=str(int(lhype_raw)),
                        slippage=0.02,
                        strategy_name="boros_hype_strategy",
                    )
                    messages.append("lHYPE swapped to HYPE")
            except Exception as e:
                messages.append(f"Spot swap error: {e}")

        if self.hyperliquid_adapter and self.balance_adapter:
            try:
                # Some routes return WHYPE instead of native HYPE. Unwrap it so we can
                # send native HYPE to Hyperliquid for liquidation.
                (
                    ok_whype,
                    whype_raw,
                ) = await self.balance_adapter.get_vault_wallet_balance(WHYPE)
                if ok_whype and whype_raw and int(whype_raw) > 0:
                    ok_unwrap, unwrap_res = await self._unwrap_whype(
                        address, int(whype_raw)
                    )
                    if ok_unwrap:
                        messages.append("WHYPE unwrapped to HYPE")
                        await asyncio.sleep(2)
                    else:
                        messages.append(f"WHYPE unwrap failed: {unwrap_res}")

                ok_hype, hype_raw = await self.balance_adapter.get_vault_wallet_balance(
                    HYPE_NATIVE
                )
                hype_raw_int = (
                    int(hype_raw) if ok_hype and hype_raw and int(hype_raw) > 0 else 0
                )
                gas_reserve_wei = int(float(MIN_HYPE_GAS) * 1e18)
                hype_to_transfer_raw = max(0, hype_raw_int - gas_reserve_wei)
                hype_to_transfer = float(hype_to_transfer_raw) / 1e18

                if hype_to_transfer_raw > int(0.01 * 1e18):
                    destination = HyperliquidAdapter.hypercore_index_to_system_address(
                        150
                    )
                    ok_send, _ = await self.balance_adapter.send_to_address(
                        token_id=HYPE_NATIVE,
                        amount=int(hype_to_transfer_raw),
                        from_wallet=strategy_wallet,
                        to_address=destination,
                        signing_callback=self._sign_callback,
                    )
                    if ok_send:
                        await asyncio.sleep(3)
                        messages.append(
                            f"HYPE transferred to HL: {hype_to_transfer:.4f}"
                        )

                        # Sell HYPE for USDC on HL spot
                        (
                            ok_spot,
                            spot_state,
                        ) = await self.hyperliquid_adapter.get_spot_user_state(address)
                        spot_hype = 0.0
                        if ok_spot and isinstance(spot_state, dict):
                            for bal in spot_state.get("balances", []):
                                if (
                                    bal.get("coin") == "HYPE"
                                    or bal.get("token") == "HYPE"
                                ):
                                    spot_hype = float(bal.get("total", 0)) - float(
                                        bal.get("hold", 0)
                                    )
                                    break

                        if spot_hype > 0.01:
                            spot_asset_id, _ = await self._get_hype_asset_ids()
                            rounded_size = (
                                self.hyperliquid_adapter.get_valid_order_size(
                                    int(spot_asset_id), spot_hype
                                )
                            )
                            if rounded_size > 0:
                                await self.hyperliquid_adapter.place_market_order(
                                    asset_id=int(spot_asset_id),
                                    is_buy=False,
                                    slippage=0.10,
                                    size=float(rounded_size),
                                    address=address,
                                    builder=self.builder_fee,
                                )
                                messages.append(
                                    f"HYPE sold for USDC: {rounded_size:.4f}"
                                )
            except Exception as e:
                messages.append(f"HYPE liquidation error: {e}")

        result_msg = f"[FAILSAFE] {reason} | {'; '.join(messages)}"
        logger.error(result_msg)
        self._failsafe_message = result_msg

        return False, result_msg

    async def _partial_trim_spot(
        self, params: dict[str, Any], inventory: Inventory
    ) -> tuple[bool, str]:
        trim_pct = float(params.get("trim_pct") or 0.25)

        if inventory.spot_value_usd < 10.0:
            return True, "No spot to trim"

        ok, msg = self._require_adapters(
            "balance_adapter", "hyperliquid_adapter", "brap_adapter"
        )
        if not ok:
            return False, msg
        if not self._sign_callback:
            return False, "No strategy wallet signing callback configured"

        ok_addr, address = self._require_strategy_wallet_address()
        if not ok_addr:
            return False, address
        strategy_wallet = self._config.get("strategy_wallet", {})

        hype_price = float(inventory.hype_price_usd or 0.0)
        if hype_price <= 0:
            ok_mid, mids = await self.hyperliquid_adapter.get_all_mid_prices()
            if ok_mid and isinstance(mids, dict):
                hype_price = float(mids.get("HYPE", 0.0))
        if hype_price <= 0:
            return False, "Could not determine HYPE price for trim"

        trim_usd = float(inventory.spot_value_usd) * float(trim_pct)

        # Sell kHYPE first (more liquid), then looped HYPE if needed.
        if inventory.khype_value_usd > 0 and trim_usd > 1.0:
            khype_trim_usd = min(float(inventory.khype_value_usd), trim_usd)
            if inventory.khype_to_hype_ratio > 0:
                khype_trim_tokens = (
                    khype_trim_usd / hype_price / float(inventory.khype_to_hype_ratio)
                )
                khype_trim_wei = int(khype_trim_tokens * 1e18)
                if khype_trim_wei > 0:
                    try:
                        await self.brap_adapter.swap_from_token_ids(
                            from_token_id=KHYPE_LST,
                            to_token_id=HYPE_NATIVE,
                            from_address=address,
                            amount=str(int(khype_trim_wei)),
                            slippage=0.01,
                            strategy_name="boros_hype_strategy",
                        )
                        trim_usd -= khype_trim_usd
                        await asyncio.sleep(2)
                    except Exception as exc:  # noqa: BLE001
                        logger.warning(f"Failed to sell kHYPE: {exc}")

        if trim_usd > 1.0 and inventory.looped_hype_value_usd > 0:
            lhype_trim_usd = min(float(inventory.looped_hype_value_usd), trim_usd)
            if inventory.looped_hype_to_hype_ratio > 0:
                lhype_trim_tokens = (
                    lhype_trim_usd
                    / hype_price
                    / float(inventory.looped_hype_to_hype_ratio)
                )
                lhype_trim_wei = int(lhype_trim_tokens * 1e18)
                if lhype_trim_wei > 0:
                    try:
                        await self.brap_adapter.swap_from_token_ids(
                            from_token_id=LOOPED_HYPE,
                            to_token_id=HYPE_NATIVE,
                            from_address=address,
                            amount=str(int(lhype_trim_wei)),
                            slippage=0.01,
                            strategy_name="boros_hype_strategy",
                        )
                        trim_usd -= lhype_trim_usd
                        await asyncio.sleep(2)
                    except Exception as exc:  # noqa: BLE001
                        logger.warning(f"Failed to sell looped HYPE: {exc}")

        remaining_trim_usd = max(0.0, float(trim_usd))
        gas_reserve_wei = int(float(MIN_HYPE_GAS) * 1e18)

        # If we still need to trim, move just enough HYPE from HyperEVM -> HL spot so we
        # can sell it for USDC and add to perp margin (avoid transferring all HYPE).
        if remaining_trim_usd > 1.0:
            # Some routes return WHYPE instead of native HYPE. Unwrap enough to cover the trim.
            try:
                ok_h0, hype_raw0 = await self.balance_adapter.get_vault_wallet_balance(
                    HYPE_NATIVE
                )
                ok_w0, whype_raw0 = await self.balance_adapter.get_vault_wallet_balance(
                    WHYPE
                )
                hype_raw0_int = (
                    int(hype_raw0) if ok_h0 and hype_raw0 and int(hype_raw0) > 0 else 0
                )
                whype_raw0_int = (
                    int(whype_raw0)
                    if ok_w0 and whype_raw0 and int(whype_raw0) > 0
                    else 0
                )
                available_native_tokens = (
                    float(max(0, hype_raw0_int - gas_reserve_wei)) / 1e18
                )
                needed_tokens = float(remaining_trim_usd) / float(hype_price)
                if (
                    whype_raw0_int > 0
                    and available_native_tokens + 1e-9 < needed_tokens
                ):
                    need_wei = int((needed_tokens - available_native_tokens) * 1e18)
                    unwrap_wei = min(int(need_wei), int(whype_raw0_int))
                    if unwrap_wei > 0:
                        ok_unwrap, unwrap_res = await self._unwrap_whype(
                            address, int(unwrap_wei)
                        )
                        if ok_unwrap:
                            await asyncio.sleep(2)
                        else:
                            logger.warning(
                                f"WHYPE unwrap failed during trim: {unwrap_res}"
                            )
            except Exception as exc:  # noqa: BLE001
                logger.debug(f"WHYPE unwrap pre-trim failed: {exc}")

            ok_hype, hype_raw = await self.balance_adapter.get_vault_wallet_balance(
                HYPE_NATIVE
            )
            hype_raw_int = (
                int(hype_raw) if ok_hype and hype_raw and int(hype_raw) > 0 else 0
            )
            transferable_raw = max(0, hype_raw_int - gas_reserve_wei)
            transferable_tokens = float(transferable_raw) / 1e18

            desired_tokens = min(transferable_tokens, remaining_trim_usd / hype_price)
            hype_to_transfer_raw = int(desired_tokens * 1e18)
            if hype_to_transfer_raw >= int(0.01 * 1e18):
                destination = HyperliquidAdapter.hypercore_index_to_system_address(150)
                ok_send, send_res = await self.balance_adapter.send_to_address(
                    token_id=HYPE_NATIVE,
                    amount=int(hype_to_transfer_raw),
                    from_wallet=strategy_wallet,
                    to_address=destination,
                    signing_callback=self._sign_callback,
                )
                if not ok_send:
                    return False, f"Failed to transfer HYPE to Hyperliquid: {send_res}"
                await asyncio.sleep(3)

        try:
            spot_asset_id, perp_asset_id = await self._get_hype_asset_ids()
            ok_spot, spot_state = await self.hyperliquid_adapter.get_spot_user_state(
                address
            )
            spot_hype_balance = 0.0
            if ok_spot and isinstance(spot_state, dict):
                for bal in spot_state.get("balances", []):
                    token = bal.get("coin") or bal.get("token")
                    hold = float(bal.get("hold", 0))
                    total = float(bal.get("total", 0))
                    available = total - hold
                    if token == "HYPE":
                        spot_hype_balance = available
                        break

            if spot_hype_balance > 0.01 and remaining_trim_usd > 1.0:
                desired_units = min(spot_hype_balance, remaining_trim_usd / hype_price)
                rounded_units = self.hyperliquid_adapter.get_valid_order_size(
                    int(spot_asset_id), desired_units
                )
                if rounded_units > 0:
                    if (
                        inventory.hl_short_size_hype > 0.1
                        and inventory.hl_short_size_hype
                        >= inventory.total_hype_exposure
                    ):
                        ok_lev, lev_msg = await self._ensure_hl_hype_leverage_set(
                            address
                        )
                        if not ok_lev:
                            return False, lev_msg
                        paired_filler = PairedFiller(
                            adapter=self.hyperliquid_adapter, address=address
                        )
                        (
                            _filled_spot,
                            _filled_perp,
                            _spot_notional,
                            _perp_notional,
                            spot_pointers,
                            perp_pointers,
                        ) = await paired_filler.fill_pair_units(
                            coin="HYPE",
                            spot_asset_id=int(spot_asset_id),
                            perp_asset_id=int(perp_asset_id),
                            total_units=float(rounded_units),
                            direction="short_spot_long_perp",
                            builder_fee=self.builder_fee,
                        )
                        await self._cancel_lingering_orders(
                            spot_pointers + perp_pointers, address
                        )
                        await asyncio.sleep(2)
                    else:
                        await self.hyperliquid_adapter.place_market_order(
                            asset_id=int(spot_asset_id),
                            is_buy=False,
                            slippage=0.10,
                            size=float(rounded_units),
                            address=address,
                            builder=self.builder_fee,
                        )
                        await asyncio.sleep(2)
                    remaining_trim_usd = max(
                        0.0, remaining_trim_usd - float(rounded_units) * hype_price
                    )
        except Exception as exc:  # noqa: BLE001
            logger.warning(f"Failed to sell spot HYPE: {exc}")

        try:
            usdc_sz_decimals = 8

            ok_spot, spot_state = await self.hyperliquid_adapter.get_spot_user_state(
                address
            )
            spot_usdc = 0.0
            spot_total_s = "0"
            spot_hold_s = "0"
            if ok_spot and isinstance(spot_state, dict):
                for bal in spot_state.get("balances", []):
                    token = bal.get("coin") or bal.get("token")
                    if token != "USDC":
                        continue
                    spot_total_s = str(bal.get("total", "0") or "0")
                    spot_hold_s = str(bal.get("hold", "0") or "0")
                    hold = float(spot_hold_s)
                    total = float(spot_total_s)
                    spot_usdc = max(0.0, total - hold)
                    break

            if spot_usdc > 1.0:
                amount = self.hyperliquid_adapter.max_transferable_amount(
                    spot_total_s,
                    spot_hold_s,
                    sz_decimals=int(usdc_sz_decimals),
                    leave_one_tick=True,
                )
                (
                    ok_xfer,
                    res_xfer,
                ) = await self.hyperliquid_adapter.transfer_spot_to_perp(
                    amount=float(amount),
                    address=address,
                )
                if (not ok_xfer) and int(usdc_sz_decimals) != 2:
                    if "insufficient balance" in str(res_xfer).lower():
                        fallback_2dp = self.hyperliquid_adapter.max_transferable_amount(
                            spot_total_s,
                            spot_hold_s,
                            sz_decimals=2,
                            leave_one_tick=True,
                        )
                        if fallback_2dp > 1.0:
                            await self.hyperliquid_adapter.transfer_spot_to_perp(
                                amount=float(fallback_2dp),
                                address=address,
                            )
        except Exception as exc:  # noqa: BLE001
            logger.warning(f"Failed to move USDC spot→perp: {exc}")

        inv_after = await self.observe()
        ok_short, msg_short = await self._ensure_hl_short(
            {
                "target_size": inv_after.total_hype_exposure,
                "current_size": inv_after.hl_short_size_hype,
            },
            inv_after,
        )
        if not ok_short:
            return False, f"Failed to resize short after trim: {msg_short}"

        return (
            True,
            f"Trimmed spot and resized short to {inv_after.total_hype_exposure:.4f} HYPE.",
        )



"""
Boros venue operations for BorosHypeStrategy.

Kept as a mixin so the main strategy file stays readable without changing behavior.
"""

from __future__ import annotations

import asyncio
from datetime import datetime
from typing import Any

from loguru import logger

from wayfinder_paths.core.utils.transaction import encode_call, send_transaction

from .constants import (
    BOROS_HYPE_MARKET_ID,
    BOROS_HYPE_TOKEN_ID,
    BOROS_MIN_DEPOSIT_HYPE,
    HYPE_NATIVE,
    HYPE_OFT_ADDRESS,
    HYPEREVM_CHAIN_ID,
    KHYPE_LST,
    LOOPED_HYPE,
    MIN_HYPE_GAS,
    USDC_ARB,
    USDT_ARB,
    WHYPE,
    WHYPE_ABI,
    WHYPE_ADDRESS,
)
from .types import Inventory


class BorosHypeBorosOpsMixin:
    async def _fund_boros(
        self, params: dict[str, Any], inventory: Inventory
    ) -> tuple[bool, str]:
        """Fund Boros using native HYPE collateral.

        Flow:
        - If we already have Arbitrum OFT HYPE, deposit it to Boros cross margin.
        - Otherwise, bridge native HYPE from HyperEVM -> Arbitrum via LayerZero OFT.
        """

        amount_usd = float(params.get("amount_usd") or 0.0)
        market_id = int(
            params.get("market_id")
            or self._planner_runtime.current_boros_market_id
            or BOROS_HYPE_MARKET_ID
        )
        token_id = int(
            params.get("token_id")
            or self._planner_runtime.current_boros_token_id
            or BOROS_HYPE_TOKEN_ID
        )

        collateral_address = str(
            params.get("collateral_address")
            or self._planner_runtime.current_boros_collateral_address
            or ""
        ).strip()
        if not collateral_address:
            collateral_address = HYPE_OFT_ADDRESS

        if amount_usd < 1.0:
            return True, f"Skipping tiny Boros funding (${amount_usd:.2f})"

        if not self._sign_callback:
            return False, "No signing callback configured"

        ok, msg = self._require_adapters("boros_adapter")
        if not ok:
            return False, msg

        ok_addr, wallet_address = self._require_strategy_wallet_address()
        if not ok_addr:
            return False, wallet_address

        hype_price = float(inventory.hype_price_usd or 0.0)
        if hype_price <= 0:
            return False, f"Invalid HYPE price (${hype_price:.6f})"

        target_hype = float(amount_usd) / hype_price
        target_wei = int(target_hype * 1e18)

        # 1) If we have OFT HYPE on Arbitrum, deposit it first (idempotent; avoids double-bridging).
        available_oft_hype = float(inventory.hype_oft_arb_balance or 0.0)
        if available_oft_hype > 0:
            deposit_hype = min(available_oft_hype, target_hype)
            deposit_usd = deposit_hype * hype_price
            if deposit_usd >= 1.0:
                deposit_wei = int(deposit_hype * 1e18)
                ok_dep, dep_res = await self.boros_adapter.deposit_to_cross_margin(
                    collateral_address=collateral_address,
                    amount_wei=deposit_wei,
                    token_id=token_id,
                    market_id=market_id,
                )
                if not ok_dep:
                    return False, f"Boros deposit failed: {dep_res}"

                logger.info(
                    f"Deposited {deposit_hype:.6f} HYPE (≈${deposit_usd:.2f}) to Boros cross margin"
                )

                # If this fully satisfies the target, stop here.
                if deposit_hype >= target_hype - 1e-9:
                    return True, (
                        f"Funded Boros with {deposit_hype:.6f} HYPE (≈${deposit_usd:.2f}) "
                        "from Arbitrum OFT balance"
                    )

                # Otherwise, bridge the remainder.
                target_hype = max(0.0, target_hype - deposit_hype)
                target_wei = int(target_hype * 1e18)
            else:
                logger.info(
                    f"Skipping tiny OFT HYPE deposit (${deposit_usd:.2f}); will top up via bridge"
                )

        # 2) Bridge native HYPE from HyperEVM to Arbitrum using the OFT contract.
        hype_balance = float(inventory.hype_hyperevm_balance or 0.0)
        whype_balance = float(inventory.whype_balance or 0.0)
        fee_buffer_hype = 0.03  # conservative buffer for OFT native fee
        desired_native_hype = float(MIN_HYPE_GAS) + float(target_hype) + fee_buffer_hype
        if whype_balance > 0.0:
            # On HyperEVM, HYPE can be split between native HYPE and WHYPE. The OFT
            # bridge requires native HYPE (msg.value), so unwrap enough WHYPE to
            # cover (amount + fee) while still leaving MIN_HYPE_GAS for future gas.
            if hype_balance < desired_native_hype:
                unwrap_hype = min(whype_balance, desired_native_hype - hype_balance)
                unwrap_wei = int(unwrap_hype * 1e18)
                if self.balance_adapter:
                    (
                        ok_whype,
                        whype_raw,
                    ) = await self.balance_adapter.get_vault_wallet_balance(WHYPE)
                    if ok_whype and isinstance(whype_raw, int) and whype_raw > 0:
                        unwrap_wei = min(int(unwrap_wei), int(whype_raw))
                if unwrap_wei > 0:
                    if not self._sign_callback:
                        return False, "No signing callback configured"

                    logger.info(
                        f"Unwrapping {unwrap_hype:.6f} WHYPE → native HYPE to fund OFT bridge"
                    )
                    try:
                        tx = await encode_call(
                            target=WHYPE_ADDRESS,
                            abi=WHYPE_ABI,
                            fn_name="withdraw",
                            args=[int(unwrap_wei)],
                            from_address=wallet_address,
                            chain_id=HYPEREVM_CHAIN_ID,
                        )
                        tx_hash = await send_transaction(
                            tx, self._sign_callback, wait_for_receipt=True
                        )
                        logger.info(f"WHYPE unwrap tx={tx_hash}")
                        await asyncio.sleep(2)
                    except Exception as exc:  # noqa: BLE001
                        logger.warning(f"WHYPE unwrap failed: {exc}")

                    if self.balance_adapter:
                        (
                            ok_hype,
                            hype_raw,
                        ) = await self.balance_adapter.get_vault_wallet_balance(
                            HYPE_NATIVE
                        )
                        if ok_hype and hype_raw:
                            hype_balance = float(hype_raw) / 1e18

        # If we still don't have enough native HYPE and there are LSTs on HyperEVM,
        # sell a small amount back to native HYPE so Boros can be funded.
        if (
            hype_balance < desired_native_hype
            and (inventory.khype_value_usd > 0 or inventory.looped_hype_value_usd > 0)
            and self.brap_adapter
        ):
            needed_hype = max(0.0, desired_native_hype - hype_balance)
            if needed_hype > 0.01:
                if inventory.khype_value_usd > 0 and inventory.khype_to_hype_ratio > 0:
                    max_hype_from_khype = float(inventory.khype_value_usd) / hype_price
                    sell_hype = min(needed_hype, max_hype_from_khype)
                    sell_khype = sell_hype / float(inventory.khype_to_hype_ratio)
                    sell_khype_wei = int(sell_khype * 1e18)
                    if sell_khype_wei > 0:
                        logger.info(
                            f"Selling {sell_khype:.6f} kHYPE → native HYPE to fund Boros"
                        )
                        ok_swap, res_swap = await self.brap_adapter.swap_from_token_ids(
                            from_token_id=KHYPE_LST,
                            to_token_id=HYPE_NATIVE,
                            from_address=wallet_address,
                            amount=str(int(sell_khype_wei)),
                            slippage=0.01,
                            strategy_name="boros_hype_strategy",
                        )
                        if ok_swap:
                            needed_hype = max(0.0, needed_hype - sell_hype)
                            await asyncio.sleep(2)
                        else:
                            logger.warning(f"Sell kHYPE→HYPE failed: {res_swap}")

                if (
                    needed_hype > 0.01
                    and inventory.looped_hype_value_usd > 0
                    and inventory.looped_hype_to_hype_ratio > 0
                ):
                    max_hype_from_lhype = (
                        float(inventory.looped_hype_value_usd) / hype_price
                    )
                    sell_hype = min(needed_hype, max_hype_from_lhype)
                    sell_lhype = sell_hype / float(inventory.looped_hype_to_hype_ratio)
                    sell_lhype_wei = int(sell_lhype * 1e18)
                    if sell_lhype_wei > 0:
                        logger.info(
                            f"Selling {sell_lhype:.6f} looped HYPE → native HYPE to fund Boros"
                        )
                        ok_swap, res_swap = await self.brap_adapter.swap_from_token_ids(
                            from_token_id=LOOPED_HYPE,
                            to_token_id=HYPE_NATIVE,
                            from_address=wallet_address,
                            amount=str(int(sell_lhype_wei)),
                            slippage=0.01,
                            strategy_name="boros_hype_strategy",
                        )
                        if ok_swap:
                            await asyncio.sleep(2)
                        else:
                            logger.warning(f"Sell looped HYPE→HYPE failed: {res_swap}")

                if self.balance_adapter:
                    (
                        ok_hype,
                        hype_raw,
                    ) = await self.balance_adapter.get_vault_wallet_balance(HYPE_NATIVE)
                    if ok_hype and hype_raw:
                        hype_balance = float(hype_raw) / 1e18
        if hype_balance <= MIN_HYPE_GAS + 0.0005:
            return (
                False,
                f"Insufficient HyperEVM HYPE to bridge (balance={hype_balance:.6f}, min_gas={MIN_HYPE_GAS:.6f})",
            )

        reserve_wei = int(MIN_HYPE_GAS * 1e18)
        balance_wei = int(hype_balance * 1e18)
        max_value_wei = max(0, balance_wei - reserve_wei)
        if max_value_wei <= 0:
            return False, "Insufficient HyperEVM HYPE to bridge after reserving gas"

        bridge_amount_wei = min(target_wei, max_value_wei)
        if bridge_amount_wei <= 0:
            return True, "Boros funding: nothing to bridge"
        (
            ok_bridge,
            bridge_res,
        ) = await self.boros_adapter.bridge_hype_oft_hyperevm_to_arbitrum(
            amount_wei=int(bridge_amount_wei),
            max_value_wei=int(max_value_wei),
            to_address=str(wallet_address),
            from_address=str(wallet_address),
        )
        if not ok_bridge:
            return False, f"OFT bridge failed: {bridge_res}"

        tx_hash = str(bridge_res.get("tx_hash") or "")
        bridged_wei = int(bridge_res.get("amount_wei") or 0)
        bridged_hype = float(bridged_wei) / 1e18
        bridged_usd = bridged_hype * hype_price

        # Track in-flight amount so planner doesn't double-fund while the bridge settles.
        self._planner_runtime.in_flight_boros_oft_hype = bridged_hype
        self._planner_runtime.in_flight_boros_oft_hype_balance_before = float(
            inventory.hype_oft_arb_balance or 0.0
        )
        self._planner_runtime.in_flight_boros_oft_hype_started_at = datetime.utcnow()

        logger.info(
            f"Initiated OFT bridge HyperEVM->Arbitrum: {bridged_hype:.6f} HYPE (≈${bridged_usd:.2f}), tx={tx_hash}"
        )

        return True, (
            f"Bridging {bridged_hype:.6f} HYPE (≈${bridged_usd:.2f}) HyperEVM→Arbitrum via OFT; "
            f"tx={tx_hash} (LayerZero: {bridge_res.get('layerzeroscan')}). "
            "Once bridged HYPE lands on Arbitrum, the next tick will deposit it to Boros."
        )

    async def _ensure_boros_position(
        self, params: dict[str, Any], inventory: Inventory
    ) -> tuple[bool, str]:
        # If Boros operations fail unexpectedly, triggers fail-safe liquidation.
        market_id = int(
            params.get("market_id")
            or self._planner_runtime.current_boros_market_id
            or BOROS_HYPE_MARKET_ID
        )
        token_id = int(
            params.get("token_id")
            or self._planner_runtime.current_boros_token_id
            or BOROS_HYPE_TOKEN_ID
        )
        target_size_yu = float(params.get("target_size_yu") or 0.0)

        if inventory.boros_pending_withdrawal_usd > 0:
            return True, (
                f"Boros withdrawal pending (~${inventory.boros_pending_withdrawal_usd:.2f}). "
                "Skipping Boros rate position actions until it settles."
            )

        if not self.boros_adapter or not market_id:
            return False, "Boros adapter not configured or no market selected"

        try:
            return await self._ensure_boros_position_impl(
                market_id=market_id,
                token_id=token_id,
                target_size_yu=target_size_yu,
                inventory=inventory,
            )
        except Exception as exc:
            logger.error(
                f"[BOROS_FAIL] Critical failure in Boros position management: {exc}"
            )
            # Trigger fail-safe liquidation
            return await self._failsafe_liquidate_all(
                f"Boros position management failed: {exc}"
            )

    async def _ensure_boros_position_impl(
        self,
        *,
        market_id: int,
        token_id: int,
        target_size_yu: float,
        inventory: Inventory,
    ) -> tuple[bool, str]:
        # Only try to manage the rate position once we have at least the minimum
        # HYPE collateral funded on Boros (or sitting idle as OFT HYPE on Arbitrum).
        depositable_hype = float(inventory.boros_collateral_hype or 0.0) + float(
            inventory.hype_oft_arb_balance or 0.0
        )
        if depositable_hype < BOROS_MIN_DEPOSIT_HYPE:
            return (
                True,
                "Skipping Boros position: collateral not funded yet "
                f"({depositable_hype:.6f} HYPE)",
            )

        # 0) Cleanup: sweep any isolated collateral back to cross margin.
        try:
            ok_sweep, sweep_res = await self.boros_adapter.sweep_isolated_to_cross(
                token_id=int(token_id)
            )
            if not ok_sweep:
                logger.warning(f"Failed Boros isolated->cross sweep: {sweep_res}")
        except Exception as exc:  # noqa: BLE001
            logger.warning(f"Failed Boros isolated->cross sweep: {exc}")

        # 1) Best-effort: if any OFT HYPE is sitting idle on Arbitrum, deposit it to cross margin.
        if inventory.hype_oft_arb_balance > 0.0:
            try:
                deposit_hype = float(inventory.hype_oft_arb_balance or 0.0)
                deposit_usd = deposit_hype * float(inventory.hype_price_usd or 0.0)
                if deposit_usd >= 1.0:
                    ok_dep, dep_res = await self.boros_adapter.deposit_to_cross_margin(
                        collateral_address=HYPE_OFT_ADDRESS,
                        amount_wei=int(deposit_hype * 1e18),
                        token_id=int(token_id),
                        market_id=int(market_id),
                    )
                    if ok_dep:
                        logger.info(
                            f"Deposited idle OFT HYPE to Boros: {deposit_hype:.6f} (≈${deposit_usd:.2f})"
                        )
                        await asyncio.sleep(2)
                    else:
                        logger.warning(
                            f"Failed to deposit OFT HYPE to Boros: {dep_res}"
                        )
            except Exception as exc:  # noqa: BLE001
                logger.warning(f"Failed to deposit OFT HYPE to Boros: {exc}")

        # 2) Rollover: close positions in other markets (best effort).
        try:
            ok_roll, roll_res = await self.boros_adapter.close_positions_except(
                keep_market_id=int(market_id),
                token_id=int(token_id),
                market_ids=inventory.boros_position_market_ids or [],
                best_effort=True,
            )
            if not ok_roll:
                logger.warning(f"Failed Boros rollover close: {roll_res}")
        except Exception as exc:  # noqa: BLE001
            logger.warning(f"Failed Boros rollover close: {exc}")

        yu_to_usd = (
            float(inventory.hype_price_usd or 0.0)
            if int(token_id) == BOROS_HYPE_TOKEN_ID
            else 1.0
        )
        ok_set, set_res = await self.boros_adapter.ensure_position_size_yu(
            market_id=int(market_id),
            token_id=int(token_id),
            target_size_yu=float(target_size_yu),
            tif="IOC",
            min_resize_excess_usd=float(
                self._planner_config.boros_resize_min_excess_usd
            ),
            yu_to_usd=float(yu_to_usd),
        )
        if not ok_set:
            return False, f"Failed to ensure Boros position: {set_res}"

        action = str(set_res.get("action") or "unknown")
        diff_yu = float(set_res.get("diff_yu") or 0.0)
        if action == "no_op":
            return (
                True,
                f"Boros position already at target ({set_res.get('current_size_yu', 0.0):.4f} YU)",
            )
        if action == "increase_short":
            return (
                True,
                f"Boros position increased by {diff_yu:.4f} YU on market {market_id}",
            )
        if action == "decrease":
            return (
                True,
                f"Boros position decreased by {abs(diff_yu):.4f} YU on market {market_id}",
            )

        return (
            True,
            f"Boros position adjusted ({action}) on market {market_id}: Δ={diff_yu:.4f} YU",
        )

    async def _complete_pending_withdrawal(
        self, params: dict[str, Any], inventory: Inventory
    ) -> tuple[bool, str]:
        # Legacy helper used by some withdrawal flows: swap USDT->USDC on Arbitrum
        usdt_idle = float(params.get("usdt_idle") or 0.0)

        if usdt_idle < 1.0:
            return True, f"Withdrawal completion: no USDT to swap (${usdt_idle:.2f})"

        ok, msg = self._require_adapters("balance_adapter", "brap_adapter")
        if not ok:
            return False, msg

        ok_addr, address = self._require_strategy_wallet_address()
        if not ok_addr:
            return False, address

        ok_usdt, usdt_raw = await self.balance_adapter.get_vault_wallet_balance(
            USDT_ARB
        )
        if not ok_usdt or usdt_raw <= 0:
            return True, f"Withdrawal completion: no USDT to swap (${usdt_idle:.2f})"

        ok_swap, swap_res = await self.brap_adapter.swap_from_token_ids(
            from_token_id=USDT_ARB,
            to_token_id=USDC_ARB,
            from_address=address,
            amount=str(int(usdt_raw)),
            slippage=0.005,
            strategy_name="boros_hype_strategy",
        )
        if not ok_swap:
            return False, f"Withdrawal completion swap failed: {swap_res}"

        ok_usdc, usdc_raw = await self.balance_adapter.get_vault_wallet_balance(
            USDC_ARB
        )
        usdc_tokens = (float(usdc_raw) / 1e6) if ok_usdc and usdc_raw > 0 else 0.0

        return True, (
            f"Withdrawal completion: swapped ${usdt_idle:.2f} USDT→USDC "
            f"(${usdc_tokens:.2f} USDC)"
        )



from __future__ import annotations

import json
from pathlib import Path

from wayfinder_paths.core.config import load_wallet_mnemonic
from wayfinder_paths.core.utils.wallets import (
    ensure_wallet_mnemonic,
    make_local_wallet,
    make_wallet_from_mnemonic,
)


def test_make_wallet_from_mnemonic_derives_metamask_addresses() -> None:
    mnemonic = "test test test test test test test test test test test junk"

    w0 = make_wallet_from_mnemonic(mnemonic, account_index=0)
    w1 = make_wallet_from_mnemonic(mnemonic, account_index=1)

    assert w0["derivation_path"] == "m/44'/60'/0'/0/0"
    assert w1["derivation_path"] == "m/44'/60'/0'/0/1"

    assert w0["address"].lower() == "0xf39fd6e51aad88f6f4ce6ab8827279cfffb92266"
    assert w1["address"].lower() == "0x70997970c51812dc3a010c7d01b50e0d17dc79c8"


def test_ensure_wallet_mnemonic_persists(tmp_path: Path) -> None:
    cfg_path = tmp_path / "config.json"
    m1 = ensure_wallet_mnemonic(config_path=cfg_path)
    m2 = ensure_wallet_mnemonic(config_path=cfg_path)

    assert isinstance(m1, str) and m1
    assert len(m1.split()) == 12
    assert m1 == m2
    assert load_wallet_mnemonic(cfg_path) == m1

    cfg = json.loads((tmp_path / "config.json").read_text())
    assert cfg["wallet_mnemonic"] == m1


def test_make_local_wallet_skips_existing_addresses_for_mnemonic() -> None:
    mnemonic = "test test test test test test test test test test test junk"
    existing = [
        {
            "label": "already_used",
            "address": make_wallet_from_mnemonic(mnemonic, account_index=1)["address"],
        }
    ]

    w = make_local_wallet(
        label="strategy_wallet",
        existing_wallets=existing,
        mnemonic=mnemonic,
    )
    assert w["derivation_index"] == 2



from typing import Any, Literal

import yaml
from pydantic import BaseModel, Field, validator


class AdapterRequirement(BaseModel):
    name: str = Field(
        ..., description="Adapter symbolic name (e.g., BALANCE, HYPERLIQUID)"
    )
    capabilities: list[str] = Field(default_factory=list)


class StrategyManifest(BaseModel):
    schema_version: str = Field(default="0.1")
    status: Literal["stable", "wip", "deprecated"] = Field(
        default="stable",
        description="Strategy maturity status. 'wip' strategies show warnings on execution.",
    )
    entrypoint: str = Field(
        ...,
        description="Python path to class, e.g. strategies.funding_rate_strategy.FundingRateStrategy",
    )
    name: str | None = Field(
        default=None,
        description="Unique name identifier for this strategy instance. Used to look up dedicated wallet in config.json by label.",
    )
    permissions: dict[str, Any] = Field(default_factory=dict)
    adapters: list[AdapterRequirement] = Field(default_factory=list)

    @validator("entrypoint")
    def validate_entrypoint(cls, v: str) -> str:
        if "." not in v:
            raise ValueError(
                "entrypoint must be a full import path to a Strategy class"
            )
        return v

    @validator("permissions")
    def validate_permissions(cls, v: dict) -> dict:
        if "policy" not in v:
            raise ValueError("permissions.policy is required")
        if not v["policy"]:
            raise ValueError("permissions.policy cannot be empty")
        return v

    @validator("adapters")
    def validate_adapters(cls, v: list) -> list:
        if not v:
            raise ValueError("adapters cannot be empty")
        return v


def load_strategy_manifest(path: str) -> StrategyManifest:
    with open(path) as f:
        data = yaml.safe_load(f)
    return StrategyManifest(**data)


def load_manifest(path: str) -> StrategyManifest:
    """Legacy function for backward compatibility."""
    return load_strategy_manifest(path)


def validate_manifest(manifest: StrategyManifest) -> None:
    # Simple v0.1 rules: require at least one adapter and permissions.policy
    if not manifest.adapters:
        raise ValueError("Manifest must declare at least one adapter")
    if "policy" not in manifest.permissions:
        raise ValueError("Manifest.permissions must include 'policy'")
